# -*- coding: utf-8 -*-
# from urllib3 because it's needed by multipart.py also in this directory
# https://github.com/shazow/urllib3/blob/4bd2f2aeec323df1c878f9f1aec69ce9d6d01fcd/urllib3/fields.py

# This is the MIT license: http://www.opensource.org/licenses/mit-license.php

# Copyright 2008-2016 Andrey Petrov and contributors (see CONTRIBUTORS.txt)

# Permission is hereby granted, free of charge, to any person obtaining a copy of this
# software and associated documentation files (the "Software"), to deal in the Software
# without restriction, including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
# FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import absolute_import
import email.utils
import mimetypes


# shim six to avoid changing the actual code from urllib3 below
class Six(object):
    import sys
    PY3 = sys.version_info[0] > 2


six = Six()


def guess_content_type(filename, default='application/octet-stream'):
    """
    Guess the "Content-Type" of a file.

    :param filename:
        The filename to guess the "Content-Type" of using :mod:`mimetypes`.
    :param default:
        If no "Content-Type" can be guessed, default to `default`.
    """
    if filename:
        return mimetypes.guess_type(filename)[0] or default
    return default


def format_header_param(name, value):
    """
    Helper function to format and quote a single header parameter.

    Particularly useful for header parameters which might contain
    non-ASCII values, like file names. This follows RFC 2231, as
    suggested by RFC 2388 Section 4.4.

    :param name:
        The name of the parameter, a string expected to be ASCII only.
    :param value:
        The value of the parameter, provided as a unicode string.
    """
    if not any(ch in value for ch in '"\\\r\n'):
        result = '%s="%s"' % (name, value)
        try:
            result.encode('ascii')
        except UnicodeEncodeError:
            pass
        else:
            return result
    if not six.PY3:  # Python 2:
        value = value.encode('utf-8')
    value = email.utils.encode_rfc2231(value, 'utf-8')
    value = '%s*=%s' % (name, value)
    return value


class RequestField(object):
    """
    A data container for request body parameters.

    :param name:
        The name of this request field.
    :param data:
        The data/value body.
    :param filename:
        An optional filename of the request field.
    :param headers:
        An optional dict-like object of headers to initially use for the field.
    """

    def __init__(self, name, data, filename=None, headers=None):
        self._name = name
        self._filename = filename
        self.data = data
        self.headers = {}
        if headers:
            self.headers = dict(headers)

    @classmethod
    def from_tuples(cls, fieldname, value):
        """
        A :class:`~urllib3.fields.RequestField` factory from old-style tuple parameters.

        Supports constructing :class:`~urllib3.fields.RequestField` from
        parameter of key/value strings AND key/filetuple. A filetuple is a
        (filename, data, MIME type) tuple where the MIME type is optional.
        For example::

            'foo': 'bar',
            'fakefile': ('foofile.txt', 'contents of foofile'),
            'realfile': ('barfile.txt', open('realfile').read()),
            'typedfile': ('bazfile.bin', open('bazfile').read(), 'image/jpeg'),
            'nonamefile': 'contents of nonamefile field',

        Field names and filenames must be unicode.
        """
        if isinstance(value, tuple):
            if len(value) == 3:
                filename, data, content_type = value
            else:
                filename, data = value
                content_type = guess_content_type(filename)
        else:
            filename = None
            content_type = None
            data = value

        request_param = cls(fieldname, data, filename=filename)
        request_param.make_multipart(content_type=content_type)

        return request_param

    def _render_part(self, name, value):
        """
        Overridable helper function to format a single header parameter.

        :param name:
            The name of the parameter, a string expected to be ASCII only.
        :param value:
            The value of the parameter, provided as a unicode string.
        """
        return format_header_param(name, value)

    def _render_parts(self, header_parts):
        """
        Helper function to format and quote a single header.

        Useful for single headers that are composed of multiple items. E.g.,
        'Content-Disposition' fields.

        :param header_parts:
            A sequence of (k, v) typles or a :class:`dict` of (k, v) to format
            as `k1="v1"; k2="v2"; ...`.
        """
        parts = []
        iterable = header_parts
        if isinstance(header_parts, dict):
            iterable = header_parts.items()

        for name, value in iterable:
            if value:
                parts.append(self._render_part(name, value))

        return '; '.join(parts)

    def render_headers(self):
        """
        Renders the headers for this request field.
        """
        lines = []

        sort_keys = ['Content-Disposition', 'Content-Type', 'Content-Location']
        for sort_key in sort_keys:
            if self.headers.get(sort_key, False):
                lines.append('%s: %s' % (sort_key, self.headers[sort_key]))

        for header_name, header_value in self.headers.items():
            if header_name not in sort_keys:
                if header_value:
                    lines.append('%s: %s' % (header_name, header_value))

        lines.append('\r\n')
        return '\r\n'.join(lines)

    def make_multipart(self, content_disposition=None, content_type=None, content_location=None):
        """
        Makes this request field into a multipart request field.

        This method overrides "Content-Disposition", "Content-Type" and
        "Content-Location" headers to the request parameter.

        :param content_type:
            The 'Content-Type' of the request body.
        :param content_location:
            The 'Content-Location' of the request body.

        """
        self.headers['Content-Disposition'] = content_disposition or 'form-data'
        self.headers['Content-Disposition'] += '; '.join(
            ['', self._render_parts((('name', self._name), ('filename', self._filename)))])
        self.headers['Content-Type'] = content_type
        self.headers['Content-Location'] = content_location



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from copy import deepcopy
import errno
import platform
import os

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.cli.run import run_command, main
from anaconda_project.internal.cli.prepare_with_mode import UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME

from anaconda_project.test.project_utils import project_dir_disable_dedicated_env


class Args(object):
    def __init__(self, **kwargs):
        self.directory = "."
        self.env_spec = None
        self.mode = UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT
        self.command = None
        self.extra_args_for_command = None
        for key in kwargs:
            setattr(self, key, kwargs[key])


python_exe = "python"
if platform.system() == 'Windows':
    python_exe = "python.exe"


def test_run_command(monkeypatch):

    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    mock_environ = deepcopy(os.environ)
    mock_environ['FOO'] = 'bar'

    monkeypatch.setattr('os.environ', mock_environ)
    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run(dirname):
        project_dir_disable_dedicated_env(dirname)

        result = run_command(
            dirname,
            UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
            conda_environment=None,
            command_name=None,
            extra_command_args=None)
        assert result is None
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert '--version' == executed['args'][1]
        assert 'bar' == executed['env']['FOO']

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME:
        """
variables:
  FOO: {}

commands:
  default:
    conda_app_entry: python --version

"""
    }, check_run)


def test_run_command_no_app_entry(capsys):
    def check_run_no_app_entry(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = run_command(
            dirname,
            UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
            conda_environment=None,
            command_name=None,
            extra_command_args=None)
        assert result is None

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: """

"""}, check_run_no_app_entry)

    out, err = capsys.readouterr()
    assert out == ""
    assert 'No known run command' in err


def test_run_command_nonexistent_project(capsys):
    def check_run_nonexistent(dirname):
        project_dir = os.path.join(dirname, "nope")
        result = _parse_args_and_run_subcommand(['anaconda-project', 'run', '--directory', project_dir])

        assert 1 == result

        out, err = capsys.readouterr()
        assert out == ""
        assert ("Project directory '%s' does not exist." % project_dir) in err

    with_directory_contents(dict(), check_run_nonexistent)


def test_run_command_failed_prepare(capsys):
    def check_run_failed_prepare(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = run_command(
            dirname,
            UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
            conda_environment=None,
            command_name=None,
            extra_command_args=None)
        assert result is None

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  - WILL_NOT_BE_SET
"""
    }, check_run_failed_prepare)

    out, err = capsys.readouterr()
    assert out == ""
    assert 'Environment variable WILL_NOT_BE_SET is not set' in err


def test_main(monkeypatch, capsys):
    def mock_conda_create(prefix, pkgs, channels):
        raise RuntimeError("this test should not create an environment in %s with pkgs %r" % (prefix, pkgs))

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = main(Args(directory=dirname))

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert '--version' == executed['args'][1]

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  default:
    conda_app_entry: python --version

"""
        }, check_run_main)

    out, err = capsys.readouterr()
    assert "" == out
    assert "" == err


def test_main_failed_exec(monkeypatch, capsys):
    def mock_execvpe(file, args, env):
        raise OSError(errno.ENOMEM, "It did not work, Michael")

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = main(Args(directory=dirname))

        assert 1 == result

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  default:
    conda_app_entry: python --version

"""
        }, check_run_main)

    out, err = capsys.readouterr()
    assert "" == out
    assert 'Failed to execute' in err
    assert 'It did not work, Michael' in err


def test_main_dirname_not_provided_use_pwd(monkeypatch, capsys):
    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(['anaconda-project', 'run'])

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert '--version' == executed['args'][1]

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  default:
    conda_app_entry: python --version

"""
        }, check_run_main)

    out, err = capsys.readouterr()
    assert "" == out
    assert "" == err


def test_run_command_extra_args(monkeypatch, capsys):
    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'run', '--directory', dirname, 'default', 'foo', '$PATH', '--something'])

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert len(executed['args']) == 5
        assert '--version' == executed['args'][1]
        assert 'foo' == executed['args'][2]
        assert '$PATH' == executed['args'][3]
        assert '--something' == executed['args'][4]

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  default:
    conda_app_entry: python --version

"""
        }, check_run_main)

    out, err = capsys.readouterr()
    assert "" == out
    assert "" == err


def test_run_command_verbose(monkeypatch, capsys):
    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', '--verbose', 'run', '--directory', dirname, 'default'])

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert len(executed['args']) == 2
        assert '--version' == executed['args'][1]

        # conda info is cached so may not be here depending on
        # which other tests run
        log_lines = ["$ conda info --json", "$ %s --version" % executed['args'][0]]
        log_lines_without_conda_info = log_lines[1:]

        def nl(lines):
            return ("\n".join(lines) + "\n")

        out, err = capsys.readouterr()
        assert "" == out
        assert nl(log_lines) == err or nl(log_lines_without_conda_info) == err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  default:
    conda_app_entry: python --version

"""
        }, check_run_main)


def test_run_command_extra_args_with_double_hyphen(monkeypatch, capsys):
    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        project_dir_disable_dedicated_env(dirname)
        # double hyphen lets us specify "--foo" as a command name
        result = _parse_args_and_run_subcommand(
            ['anaconda-project', 'run', '--directory', dirname, '--', '--foo', '--bar'])

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        assert executed['file'].endswith(python_exe)
        assert executed['args'][0].endswith(python_exe)
        assert len(executed['args']) == 3
        assert '--version' == executed['args'][1]
        assert '--bar' == executed['args'][2]

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
commands:
  "--foo":
    conda_app_entry: python --version
"""
        }, check_run_main)

    out, err = capsys.readouterr()
    assert "" == out
    assert "" == err


def _is_python_exe(path):
    assert path.endswith(python_exe)


def _test_run_command_foo(command_line, monkeypatch, capsys, file_assertion=_is_python_exe):
    executed = {}

    def mock_execvpe(file, args, env):
        executed['file'] = file
        executed['args'] = args
        executed['env'] = env

    monkeypatch.setattr('os.execvpe', mock_execvpe)

    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        project_dir_disable_dedicated_env(dirname)

        for n, i in enumerate(command_line):
            if i == '<DIRNAME>':
                command_line[n] = dirname

        result = _parse_args_and_run_subcommand(command_line)

        assert 1 == result
        assert 'file' in executed
        assert 'args' in executed
        assert 'env' in executed
        file_assertion(executed['file'])

        out, err = capsys.readouterr()
        assert "" == out
        assert "" == err

        return executed['args'][1:]

    return with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME:
        """
commands:
  default:
    conda_app_entry: python --version def
  foo:
    conda_app_entry: python --version foo
  bar:
    conda_app_entry: python --version bar
"""
    }, check_run_main)


def test_run_command_specify_name_after_options(monkeypatch, capsys):
    args = _test_run_command_foo(['anaconda-project', 'run', '--directory', '<DIRNAME>', 'foo'], monkeypatch, capsys)

    assert args == ['--version', 'foo']


def test_run_command_specify_name_before_options(monkeypatch, capsys):
    args = _test_run_command_foo(['anaconda-project', 'run', 'foo', '--directory', '<DIRNAME>'], monkeypatch, capsys)
    assert args[:-1] == ['--version', 'foo', '--directory']


def test_run_command_omit_name_use_default(monkeypatch, capsys):
    args = _test_run_command_foo(['anaconda-project', 'run', '--directory', '<DIRNAME>'], monkeypatch, capsys)
    assert args == ['--version', 'def']


# can't put an assert in a lambda so this makes us a "lambda" with
# an assert in it
def _func_asserting_contains(what):
    def f(s):
        assert what in s

    return f


def test_run_command_executable_not_in_config(monkeypatch, capsys):
    args = _test_run_command_foo(['anaconda-project', 'run', '--directory', '<DIRNAME>', 'something1', 'something2'],
                                 monkeypatch, capsys, _func_asserting_contains('something1'))
    assert args == ['something2']


def test_run_notebook_not_in_config(monkeypatch, capsys):
    args = _test_run_command_foo(['anaconda-project', 'run', '--directory', '<DIRNAME>', 'something.ipynb'],
                                 monkeypatch, capsys, _func_asserting_contains('jupyter-notebook'))
    assert len(args) == 2
    assert args[0].endswith('something.ipynb')
    assert args[1] == '--NotebookApp.default_url=/notebooks/something.ipynb'


def test_run_command_nonexistent_name(monkeypatch, capsys):
    def check_run_main(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)

        def mock_execvpe(file, args, env):
            assert file == 'nope'
            assert args == ['nope']
            raise OSError("no such file nope")

        monkeypatch.setattr('os.execvpe', mock_execvpe)

        project_dir_disable_dedicated_env(dirname)
        result = _parse_args_and_run_subcommand(['anaconda-project', 'run', '--directory', dirname, 'nope'])

        assert 1 == result

        out, err = capsys.readouterr()
        assert "" == out
        assert "Failed to execute 'nope'" in err

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME:
        """
commands:
  default:
    conda_app_entry: python --version
  foo:
    conda_app_entry: python --version foo
  bar:
    conda_app_entry: python --version bar
"""
    }, check_run_main)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Project class representing a project directory."""
from __future__ import absolute_import

import contextlib
from copy import deepcopy, copy
import os
from os.path import join

from anaconda_project.env_spec import (EnvSpec, _anaconda_default_env_spec, _find_importable_spec,
                                       _find_out_of_sync_importable_spec)
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import EnvVarRequirement
from anaconda_project.requirements_registry.requirements.conda_env import (CondaEnvRequirement,
                                                                           CondaBootstrapEnvRequirement)
from anaconda_project.requirements_registry.requirements.download import DownloadRequirement
from anaconda_project.requirements_registry.requirements.service import ServiceRequirement
from anaconda_project.project_commands import (ProjectCommand, all_known_command_attributes)
from anaconda_project.project_file import ProjectFile
from anaconda_project.project_lock_file import ProjectLockFile
from anaconda_project.archiver import _list_relative_paths_for_unignored_project_files
from anaconda_project.version import version
from anaconda_project.conda_manager import CondaLockSet
from anaconda_project.frontend import _null_frontend, _new_error_recorder, Frontend
from anaconda_project.yaml_file import CommentedMap

from anaconda_project.internal.py2_compat import is_string, is_list, is_dict
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.internal.slugify import slugify
import anaconda_project.internal.notebook_analyzer as notebook_analyzer
import anaconda_project.internal.conda_api as conda_api
import anaconda_project.internal.pip_api as pip_api
from anaconda_project.internal import plugins as plugins_api

# These strings are used in the command line options to anaconda-project,
# so changing them has back-compat consequences.
COMMAND_TYPE_CONDA_APP_ENTRY = 'conda_app_entry'
COMMAND_TYPE_SHELL = 'unix'
COMMAND_TYPE_WINDOWS = 'windows'
COMMAND_TYPE_NOTEBOOK = 'notebook'
COMMAND_TYPE_BOKEH_APP = 'bokeh_app'

ALL_COMMAND_TYPES = (COMMAND_TYPE_CONDA_APP_ENTRY, COMMAND_TYPE_SHELL, COMMAND_TYPE_WINDOWS, COMMAND_TYPE_NOTEBOOK,
                     COMMAND_TYPE_BOKEH_APP)


class ProjectProblem(object):
    """A possibly-autofixable problem with a project."""

    def __init__(self,
                 text,
                 filename=None,
                 fix_prompt=None,
                 fix_function=None,
                 no_fix_function=None,
                 only_a_suggestion=False,
                 line_number=None,
                 column_number=None):
        """Create a project problem."""
        self.text_without_filename = text
        if filename is None:
            self.text = text
        else:
            self.text = "%s: %s" % (os.path.basename(filename), text)
        self.fix_prompt = fix_prompt
        self.fix_function = fix_function
        self.no_fix_function = no_fix_function
        self.only_a_suggestion = only_a_suggestion
        self.maybe_filename = filename
        self.maybe_line_number = line_number
        self.maybe_column_number = column_number

    @property
    def can_fix(self):
        """True if the problem can be auto-fixed."""
        return self.fix_function is not None

    def fix(self, project):
        """Perform the auto-fix."""
        if self.fix_function is not None:
            return self.fix_function(project)
        else:
            return None

    def no_fix(self, project):
        """Take an action on deciding not to fix."""
        if self.no_fix_function is not None:
            return self.no_fix_function(project)
        else:
            return None


# given a list of mixed strings and ProjectProblem, make
# them all into ProjectProblem
def _make_problems_into_objects(problems):
    new_problems = []
    for p in problems:
        if isinstance(p, ProjectProblem):
            new_problems.append(p)
        else:
            new_problems.append(ProjectProblem(text=p))
    return new_problems


def _file_problem(problems, yaml_file, text, fix_prompt=None, fix_function=None):
    problems.append(
        ProjectProblem(text=text, filename=yaml_file.filename, fix_prompt=fix_prompt, fix_function=fix_function))


def _unknown_field_suggestions(project_file, problems, yaml_dict, known_fields):
    if 'user_fields' in yaml_dict.keys():
        known_fields = _add_user_fields(yaml_dict, known_fields)
    for key in yaml_dict.keys():
        if key not in known_fields:
            problems.append(
                ProjectProblem(
                    text="Unknown field name '%s'" % (key), filename=project_file.filename, only_a_suggestion=True))


def _add_user_fields(yaml_dict, known_fields):
    user_fields = yaml_dict['user_fields']
    user_fields.append('user_fields')
    all = [known_fields, user_fields]
    return [i for sub in all for i in sub]


def _fatal_problem(problems):
    for p in problems:
        # strings are fatal problems
        if not isinstance(p, ProjectProblem):
            return True
        # ProjectProblem instances may be fatal problems
        if not p.only_a_suggestion:
            return True
    return False


class _ConfigCache(object):
    def __init__(self, directory_path, registry, must_exist):
        self.directory_path = directory_path
        if registry is None:
            registry = RequirementsRegistry()
        self.registry = registry

        self.name = None
        self.description = ''
        self.icon = None
        self.commands = dict()
        self.default_command_name = None
        self.project_file_count = 0
        self.lock_file_count = 0
        self.env_specs = dict()
        self.lock_sets = dict()
        self.locking_globally_enabled = False
        self.default_env_spec_name = None
        self.global_base_env_spec = None
        self.must_exist = must_exist

    def update(self, project_file, lock_file):
        if project_file.change_count == self.project_file_count and \
                lock_file.change_count == self.lock_file_count:
            return

        self.project_file_count = project_file.change_count
        self.lock_file_count = lock_file.change_count

        requirements = dict()
        problems = []

        def accept_project_creation(project):
            self.must_exist = False

        project_exists = os.path.isdir(self.directory_path)
        if not project_exists:
            problems.append("Project directory '%s' does not exist." % self.directory_path)
        elif self.must_exist and not os.path.isfile(project_file.filename):
            problems.append(
                ProjectProblem(
                    text="Project file '%s' does not exist." % os.path.basename(project_file.filename),
                    fix_prompt="Create file '%s'?" % project_file.filename,
                    fix_function=accept_project_creation))

        if project_file.corrupted:
            problems.append(
                ProjectProblem(
                    text=("Syntax error: %s" % (project_file.corrupted_error_message)),
                    filename=project_file.filename,
                    line_number=project_file.corrupted_maybe_line,
                    column_number=project_file.corrupted_maybe_column))

        if lock_file.corrupted:
            problems.append(
                ProjectProblem(
                    text=("Syntax error: %s" % (lock_file.corrupted_error_message)),
                    filename=lock_file.filename,
                    line_number=lock_file.corrupted_maybe_line,
                    column_number=lock_file.corrupted_maybe_column))

        if project_exists and not (project_file.corrupted or lock_file.corrupted):
            _unknown_field_suggestions(
                project_file, problems, project_file.root,
                ('name', 'description', 'icon', 'variables', 'downloads', 'services', 'env_specs', 'commands',
                 'packages', 'dependencies', 'channels', 'platforms', 'skip_imports'))

            _unknown_field_suggestions(lock_file, problems, lock_file.root, ('env_specs', 'locking_enabled'))

            self._update_name(problems, project_file)
            self._update_description(problems, project_file)
            self._update_icon(problems, project_file)
            self._update_lock_sets(problems, lock_file)
            self._update_env_specs(problems, project_file, lock_file)
            # future: we could un-hardcode this so plugins can add stuff here
            self._update_variables(requirements, problems, project_file)
            self._update_downloads(requirements, problems, project_file)
            self._update_services(requirements, problems, project_file)
            # this MUST be after we _update_variables since we may get CondaEnvRequirement
            # options in the variables section, and after _update_env_specs
            # since we use those
            self._update_conda_env_requirements(requirements, problems, project_file)

            # this MUST be after we update env reqs so we have the valid env spec names
            self._update_commands(problems, project_file, requirements)

            self._verify_command_dependencies(problems, project_file)

        self.requirements = requirements
        self.problems = _make_problems_into_objects(problems)
        self.problem_strings = list([p.text for p in self.problems if not p.only_a_suggestion])

    def _update_name(self, problems, project_file):
        # For back-compat reasons, name=null means auto-name at runtime,
        # while name field missing entirely is an error.
        default_name = os.path.basename(self.directory_path)

        if 'name' not in project_file.root:

            def set_name_field(project):
                project.project_file.set_value('name', default_name)

            problems.append(
                ProjectProblem(
                    text="The 'name:' field is missing.",
                    filename=project_file.filename,
                    fix_prompt=("Name the project '%s'?" % default_name),
                    fix_function=set_name_field))
            # Note: we continue on here to set set the default name below,
            # just to avoid dealing with `project.name is None` elsewhere
            # in the code, but we don't save the name to the project_file.

        name = project_file.get_value('name', None)
        if name is not None:
            if not is_string(name):
                _file_problem(problems, project_file, "name: field should have a string value not %r" % name)
                name = None
            elif len(name.strip()) == 0:
                _file_problem(problems, project_file, "name: field is an empty or all-whitespace string.")
                name = None

        if name is None:
            name = default_name

        self.name = name

    def _update_description(self, problems, project_file):
        desc = project_file.get_value('description', None)
        if desc is not None and not is_string(desc):
            _file_problem(problems, project_file, "description: field should have a string value not %r" % desc)
            desc = None

        if desc is None:
            desc = ''

        self.description = desc

    def _update_icon(self, problems, project_file):
        icon = project_file.get_value('icon', None)
        if icon is not None and not is_string(icon):
            _file_problem(problems, project_file, "icon: field should have a string value not %r" % (icon))
            icon = None

        if icon is not None:
            icon = os.path.join(self.directory_path, icon)
            if not os.path.isfile(icon):
                problems.append("Icon file %s does not exist." % icon)
                icon = None

        self.icon = icon

    def _add_requirement(self, requirements, env_spec, requirement):
        # note that env_spec.name is None for the global_base_env_spec
        if env_spec.name not in requirements:
            requirements[env_spec.name] = []
        requirements[env_spec.name].append(requirement)

    def _update_requirements(self, requirements, problems, project_file, dict_name, updater):
        global_dict = project_file.get_value(dict_name)
        updater(requirements, problems, project_file, self.global_base_env_spec, global_dict)
        for env_spec in self.env_specs.values():
            env_dict = project_file.get_value(['env_specs', env_spec.name, dict_name], None)
            updater(requirements, problems, project_file, env_spec, env_dict)

    def _update_variables(self, requirements, problems, project_file):
        self._update_requirements(requirements, problems, project_file, 'variables',
                                  self._update_variables_for_env_spec)

    def _update_downloads(self, requirements, problems, project_file):
        self._update_requirements(requirements, problems, project_file, 'downloads',
                                  self._update_downloads_for_env_spec)

    def _update_services(self, requirements, problems, project_file):
        self._update_requirements(requirements, problems, project_file, 'services', self._update_services_for_env_spec)

    def _update_variables_for_env_spec(self, requirements, problems, project_file, env_spec, variables):
        def check_conda_reserved(key):
            if key in ('CONDA_DEFAULT_ENV', 'CONDA_ENV_PATH', 'CONDA_PREFIX'):
                _file_problem(problems, project_file, ("Environment variable %s is reserved for Conda's use, " +
                                                       "so it can't appear in the variables section.") % key)
                return True
            else:
                return False

        # variables: section can contain a list of var names or a dict from
        # var names to options OR default values. it can also be missing
        # entirely which is the same as empty.
        if variables is None:
            pass
        elif is_dict(variables):
            for key in variables.keys():
                if check_conda_reserved(key):
                    continue
                if key.strip() == '':
                    _file_problem(problems, project_file,
                                  "Variable name cannot be empty string, found: '{}' as name".format(key))
                    continue
                raw_options = variables[key]

                if raw_options is None:
                    options = {}
                elif is_dict(raw_options):
                    options = deepcopy(raw_options)  # so we can modify it below
                else:
                    options = dict(default=raw_options)

                assert (isinstance(options, dict))

                if EnvVarRequirement._parse_default(options, key, problems):
                    requirement = self.registry.find_requirement_by_env_var(key, options)
                    self._add_requirement(requirements, env_spec, requirement)

        elif is_list(variables):
            for item in variables:
                if is_string(item):
                    if item.strip() == '':
                        _file_problem(problems, project_file,
                                      "Variable name cannot be empty string, found: '{}' as name".format(item))
                        continue
                    if check_conda_reserved(item):
                        continue

                    requirement = self.registry.find_requirement_by_env_var(item, options=dict())
                    self._add_requirement(requirements, env_spec, requirement)
                else:
                    _file_problem(
                        problems, project_file,
                        ("variables section should contain environment variable names, {item} is not a string".format(
                            item=item)))
        else:
            _file_problem(
                problems, project_file,
                "variables section contains wrong value type {value}, should be dict or list of requirements".format(
                    value=variables))

    def _update_downloads_for_env_spec(self, requirements, problems, project_file, env_spec, downloads):
        if downloads is None:
            return

        if not is_dict(downloads):
            _file_problem(problems, project_file, "'downloads:' section should be a dictionary, found {}".format(
                repr(downloads)))
            return

        for varname, item in downloads.items():
            if varname.strip() == '':
                _file_problem(problems, project_file,
                              "Download name cannot be empty string, found: '{}' as name".format(varname))
                continue
            download_kwargs = DownloadRequirement._parse(varname, item, problems)
            if download_kwargs is None:
                continue

            requirement = DownloadRequirement(self.registry, **download_kwargs)
            self._add_requirement(requirements, env_spec, requirement)

    def _update_services_for_env_spec(self, requirements, problems, project_file, env_spec, services):
        if services is None:
            return

        if not is_dict(services):
            _file_problem(problems, project_file,
                          ("'services:' section should be a dictionary from environment variable to " +
                           "service type, found {}").format(repr(services)))
            return

        for varname, item in services.items():
            if varname.strip() == '':
                _file_problem(problems, project_file,
                              "Service name cannot be empty string, found: '{}' as name".format(varname))
                continue

            service_kwargs = ServiceRequirement._parse(varname, item, problems)
            if service_kwargs is None:
                continue
            service_type = service_kwargs['service_type']

            if not self.registry.can_find_requirement_by_service_type(**service_kwargs):
                problems.append("Service {} has an unknown type '{}'.".format(varname, service_type))
                continue

            requirement = self.registry.find_requirement_by_service_type(**service_kwargs)
            assert isinstance(requirement, ServiceRequirement)
            assert 'type' in requirement.options
            self._add_requirement(requirements, env_spec, requirement)

    def _parse_string_list_with_special(self, problems, yaml_file, parent_dict, key, what, special_filter):
        items = parent_dict.get(key, [])
        if not is_list(items):
            _file_problem(problems, yaml_file, "%s: value should be a list of %ss, not '%r'" % (key, what, items))
            return ([], [])
        cleaned = []
        special = []
        for item in items:
            if is_string(item):
                cleaned.append(item.strip())
            elif special_filter(item):
                special.append(item)
            else:
                _file_problem(problems, yaml_file,
                              ("%s: value should be a %s (as a string) not '%r'" % (key, what, item)))
        return (cleaned, special)

    def _parse_string_list(self, problems, yaml_file, parent_dict, key, what):
        return self._parse_string_list_with_special(
            problems, yaml_file, parent_dict, key, what, special_filter=lambda x: False)[0]

    def _parse_platforms(self, problems, yaml_file, parent_dict):
        platforms = self._parse_string_list(problems, yaml_file, parent_dict, 'platforms', 'platform name')
        (platforms, unknown, invalid) = conda_api.validate_platform_list(platforms)
        for u in unknown:
            problems.append(
                ProjectProblem(
                    text=(
                        "Unusual platform name '%s' may be a typo (more usual examples: linux-64, osx-64, win-64)" % u),
                    filename=yaml_file.filename,
                    only_a_suggestion=True))
        for i in invalid:
            _file_problem(problems, yaml_file,
                          "Platform name '%s' is invalid (valid examples: linux-64, osx-64, win-64)" % i)
        return platforms

    def _parse_packages(self, problems, yaml_file, key, parent_dict):
        (deps, pip_dicts) = self._parse_string_list_with_special(problems, yaml_file, parent_dict, key, 'package name',
                                                                 lambda x: is_dict(x) and ('pip' in x))
        for dep in deps:
            parsed = conda_api.parse_spec(dep)
            if parsed is None:
                _file_problem(problems, yaml_file, "invalid package specification: %s" % (dep))

        # note that multiple "pip:" dicts are allowed
        pip_deps = []
        for pip_dict in pip_dicts:
            pip_list = self._parse_string_list(problems, yaml_file, pip_dict, 'pip', 'pip package name')
            pip_deps.extend(pip_list)

        for dep in pip_deps:
            parsed = pip_api.parse_spec(dep)
            if parsed is None:
                _file_problem(problems, yaml_file, "invalid pip package specifier: %s" % (dep))

        return (deps, pip_deps)

    def _update_lock_sets(self, problems, lock_file):
        self.lock_sets = dict()
        self.locking_globally_enabled = False

        enabled = lock_file.get_value(['locking_enabled'], True)
        if not isinstance(enabled, bool):
            _file_problem(problems, lock_file, "Value for locking_enabled should be true or false, found %r" % enabled)
        else:
            self.locking_globally_enabled = enabled

        lock_sets = lock_file.get_value(['env_specs'], {})
        if not is_dict(lock_sets):
            _file_problem(problems, lock_file, ("'env_specs:' section in lock file should be a dictionary from " +
                                                "env spec names to lock information, found {}").format(repr(lock_sets)))
            return

        for (name, lock_set) in lock_sets.items():
            if not is_dict(lock_set):
                _file_problem(
                    problems, lock_file,
                    "Field '%s' in env_specs in lock file should be a dictionary, found %r" % (name, lock_set))
                continue

            _unknown_field_suggestions(lock_file, problems, lock_set,
                                       ('packages', 'dependencies', 'platforms', 'locked', 'env_spec_hash'))

            enabled = lock_set.get('locked', self.locking_globally_enabled)
            if not isinstance(enabled, bool):
                _file_problem(problems, lock_file,
                              "Value for locked for env spec '%s' should be true or false, found %r" % (name, enabled))
                continue

            env_spec_hash = lock_set.get('env_spec_hash', None)
            # we deliberately don't check the hash length or format, because we might
            # want to evolve those someday, and if someone sets the hash by hand to
            # "foobar" that's fine, won't hurt anything.
            if env_spec_hash is not None and not is_string(env_spec_hash):
                _file_problem(
                    problems, lock_file,
                    "Value for env_spec_hash for env spec '%s' should be a string, found %r" % (name, env_spec_hash))
                continue

            platforms = self._parse_platforms(problems, lock_file, lock_set)

            conda_packages_by_platform = dict()
            packages_by_platform = lock_set.get('packages', {})
            if not is_dict(packages_by_platform):
                _file_problem(
                    problems, lock_file,
                    "'packages:' section in env spec '%s' in lock file should be a dictionary, found %r" %
                    (name, packages_by_platform))
                continue

            for platform in packages_by_platform.keys():
                previous_problem_count = len(problems)
                # this may set problems due to invalid package specs
                (deps, pip_deps) = self._parse_packages(problems, lock_file, platform, packages_by_platform)

                if len(problems) > previous_problem_count:
                    continue

                if len(pip_deps) > 0:
                    # we warn but don't fail on this, so if we add pip support in the future
                    # older versions of anaconda-project won't puke on it.
                    problems.append(
                        ProjectProblem(
                            text="env spec '%s': pip dependencies are currently ignored in the lock file" % name,
                            filename=lock_file.filename,
                            only_a_suggestion=True))

                conda_packages_by_platform[platform] = deps

            lock_set_object = CondaLockSet(
                package_specs_by_platform=conda_packages_by_platform, platforms=platforms, enabled=enabled)
            lock_set_object.env_spec_hash = env_spec_hash

            self.lock_sets[name] = lock_set_object

    def _update_env_specs(self, problems, project_file, lock_file):
        def _parse_string_list(parent_dict, key, what):
            return self._parse_string_list(problems, project_file, parent_dict, key, what)

        def _parse_channels(parent_dict):
            return _parse_string_list(parent_dict, 'channels', 'channel name')

        def _parse_platforms(parent_dict):
            return self._parse_platforms(problems, project_file, parent_dict)

        def _parse_packages(parent_dict):
            return self._parse_packages(problems, project_file, 'packages', parent_dict)

        (shared_deps, shared_pip_deps) = _parse_packages(project_file.root)
        shared_channels = _parse_channels(project_file.root)
        shared_platforms = _parse_platforms(project_file.root)

        _default_env_spec = CommentedMap([('default', CommentedMap([('packages', []), ('channels', [])]))])
        env_specs = project_file.get_value('env_specs', default=_default_env_spec)

        first_env_spec_name = None
        env_specs_is_empty = False
        env_specs_is_missing = False

        # this one isn't in the env_specs dict
        self.global_base_env_spec = EnvSpec(
            name=None,
            conda_packages=shared_deps,
            pip_packages=shared_pip_deps,
            channels=shared_channels,
            platforms=shared_platforms,
            description="Global packages and channels",
            inherit_from_names=(),
            inherit_from=())

        env_spec_attrs = dict()
        if env_specs is None:
            env_specs_is_missing = True
        elif is_dict(env_specs):
            if len(env_specs) == 0:
                env_specs_is_empty = True
            for (name, attrs) in env_specs.items():
                if name.strip() == '':
                    _file_problem(problems, project_file,
                                  "Environment spec name cannot be empty string, found: '{}' as name".format(name))
                    continue
                description = attrs.get('description', None)
                if description is not None and not is_string(description):
                    _file_problem(problems, project_file,
                                  ("'description' field of environment {} must be a string".format(name)))
                    continue

                problem_count = len(problems)
                inherit_from_names = attrs.get('inherit_from', None)
                if inherit_from_names is None:
                    inherit_from_names = []
                elif is_string(inherit_from_names):
                    inherit_from_names = [inherit_from_names.strip()]
                else:
                    inherit_from_names = _parse_string_list(attrs, 'inherit_from', 'env spec name')

                if len(problems) > problem_count:
                    # we got a new problem from the bad inherit_from
                    continue

                (deps, pip_deps) = _parse_packages(attrs)
                channels = _parse_channels(attrs)
                platforms = _parse_platforms(attrs)

                lock_set = self.lock_sets.get(name, None)
                if lock_set is None:
                    lock_set = CondaLockSet(
                        package_specs_by_platform=dict(),
                        platforms=[],
                        enabled=self.locking_globally_enabled,
                        missing=True)

                env_spec_attrs[name] = dict(
                    name=name,
                    conda_packages=deps,
                    pip_packages=pip_deps,
                    channels=channels,
                    platforms=platforms,
                    description=description,
                    inherit_from_names=tuple(inherit_from_names),
                    inherit_from=(),
                    lock_set=lock_set)

                if first_env_spec_name is None:
                    first_env_spec_name = name

                _unknown_field_suggestions(project_file, problems, attrs,
                                           ('packages', 'dependencies', 'channels', 'platforms', 'description',
                                            'inherit_from', 'variables', 'services', 'downloads'))
        else:
            _file_problem(
                problems, project_file,
                "env_specs should be a dictionary from environment name to environment attributes, not %r" %
                (env_specs))

        self.env_specs = dict()

        def make_env_spec(name, trail):
            assert name in env_spec_attrs

            if name not in self.env_specs:
                was_cycle = False
                if name in trail:
                    _file_problem(problems, project_file,
                                  ("'inherit_from' fields create circular inheritance among these env specs: {}".format(
                                      ", ".join(sorted(trail)))))
                    was_cycle = True
                trail.append(name)

                attrs = env_spec_attrs[name]

                if not was_cycle:
                    inherit_from_names = attrs['inherit_from_names']
                    for parent in inherit_from_names:
                        if parent not in env_spec_attrs:
                            _file_problem(problems, project_file,
                                          ("name '{}' in 'inherit_from' field of env spec {} does not match " +
                                           "the name of another env spec").format(parent, attrs['name']))
                        else:
                            inherit_from = make_env_spec(parent, trail)
                            attrs['inherit_from'] = attrs['inherit_from'] + (inherit_from, )

                # All parent-less env specs get the global base spec as parent,
                # which means the global base spec is in everyone's ancestry
                if attrs['inherit_from'] == ():
                    attrs['inherit_from'] = (self.global_base_env_spec, )

                self.env_specs[name] = EnvSpec(**attrs)

            return self.env_specs[name]

        # it's important to create all the env specs when possible
        # even if they are broken (e.g. bad inherit_from), so they
        # can be edited in order to fix them

        for name in env_spec_attrs.keys():
            make_env_spec(name, [])
            assert name in self.env_specs

        # Find issues with missing platforms: lists in the project file

        missing_platforms = []
        locked_specs_count = 0
        for env_spec in self.env_specs.values():
            if env_spec.lock_set.enabled:
                locked_specs_count += 1
            if env_spec.lock_set.enabled and len(env_spec.platforms) == 0:
                missing_platforms.append(env_spec.name)

        if locked_specs_count > 0:
            default_platforms = conda_api.default_platforms_with_current()

            # If none of the env specs have platforms, assume we want to
            # add platforms: to the toplevel (spanning entire file).
            # Otherwise, suggest fixing them one-by-one.
            missing_platform_count = len(missing_platforms)
            if missing_platform_count == locked_specs_count:

                def set_global_default_platforms(project):
                    project.project_file.set_value(['platforms'], default_platforms)

                _file_problem(
                    problems,
                    project_file,
                    "The 'platforms:' field should list platforms the project supports.",
                    fix_prompt=("Set platforms to '%s'?" % ", ".join(default_platforms)),
                    fix_function=set_global_default_platforms)
            else:
                for missing in sorted(missing_platforms):

                    def make_fix(missing):
                        def set_env_spec_platforms(project):
                            project.project_file.set_value(['env_specs', missing, 'platforms'], default_platforms)

                        return set_env_spec_platforms

                    _file_problem(
                        problems,
                        project_file,
                        "Env spec %s does not have anything in its 'platforms:' field." % missing,
                        fix_prompt=("Set platforms to '%s'?" % ", ".join(default_platforms)),
                        fix_function=make_fix(missing))

        # Find lock-set-out-of-sync-with-env-spec issues

        for env_spec in self.env_specs.values():
            if env_spec.lock_set.disabled:
                continue

            locked_hash = env_spec.lock_set.env_spec_hash
            if locked_hash is not None and locked_hash != env_spec.logical_hash:
                text = ("Env spec '%s' has changed since the lock file was last updated "
                        "(env spec hash has changed from %s to %s)") % (env_spec.name, locked_hash,
                                                                        env_spec.logical_hash)
                problems.append(ProjectProblem(text=text, filename=lock_file.filename, only_a_suggestion=True))

            if env_spec.platforms != env_spec.lock_set.platforms:
                if len(env_spec.lock_set.platforms) == 0:
                    text = "Env spec '%s' specifies platforms '%s' but the lock file lists no platforms for it" % (
                        env_spec.name, ",".join(env_spec.platforms))
                else:
                    text = ("Env spec '%s' specifies platforms '%s' but the lock file has " +
                            "locked versions for platforms '%s'") % (env_spec.name, ",".join(env_spec.platforms),
                                                                     ",".join(env_spec.lock_set.platforms))
                problems.append(ProjectProblem(text=text, filename=lock_file.filename, only_a_suggestion=True))

            if len(env_spec.conda_packages) > 0:
                for platform in env_spec.lock_set.platforms:
                    conda_packages = env_spec.lock_set.package_specs_for_platform(platform)
                    if len(conda_packages) == 0:
                        text = ("Lock file lists no packages for env spec '%s' on platform %s") % (env_spec.name,
                                                                                                   platform)
                        problems.append(ProjectProblem(text=text, filename=lock_file.filename, only_a_suggestion=True))
                    else:
                        # If conda ever had RPM-like "Obsoletes" then this situation _may_ happen
                        # in correct scenarios.
                        lock_set_names = set()
                        for package in conda_packages:
                            parsed = conda_api.parse_spec(package)
                            if parsed is not None:
                                lock_set_names.add(parsed.name)
                        unlocked_names = env_spec.conda_package_names_set - lock_set_names
                        if len(unlocked_names) > 0:
                            text = "Lock file is missing %s packages for env spec %s on %s (%s)" % (
                                len(unlocked_names), env_spec.name, platform, ",".join(sorted(list(unlocked_names))))
                            problems.append(
                                ProjectProblem(text=text, filename=lock_file.filename, only_a_suggestion=True))

        # Look for lock sets that don't go with an env spec
        for name in self.lock_sets.keys():
            if name not in self.env_specs:
                text = ("Lock file lists env spec '%s' which is not in %s") % (name, project_file.basename)
                problems.append(ProjectProblem(text=text, filename=lock_file.filename, only_a_suggestion=True))

        # Look for environment.yml, requirements.txt that are out of sync

        (importable_spec, importable_filename) = _find_out_of_sync_importable_spec(self.env_specs.values(),
                                                                                   self.directory_path)
        if importable_spec is not None:
            skip_spec_import = project_file.get_value(['skip_imports', 'environment'])
            if skip_spec_import == importable_spec.logical_hash:
                importable_spec = None

        if importable_spec is not None:
            old = self.env_specs.get(importable_spec.name)

        # this is a pretty bad hack, but if we injected "notebook"
        # or "bokeh" deps to make a notebook/bokeh command work,
        # we will end up out-of-sync for that reason
        # alone. environment.yml seems to typically not have
        # "notebook" in it because the environment.yml is used for
        # the kernel but not Jupyter itself.
        # We then end up in a problem loop where we complain about
        # missing notebook dep, add it, then complain about environment.yml
        # out of sync, and `anaconda-project init` in a directory with a .ipynb
        # and an environment.yml doesn't result in a valid project.
        if importable_spec is not None and old is not None and \
                importable_spec.diff_only_removes_notebook_or_bokeh(old):
            importable_spec = None

        if importable_spec is not None:
            if old is None:
                text = "Environment spec '%s' from %s is not in %s." % (importable_spec.name, importable_filename,
                                                                        os.path.basename(project_file.filename))
                prompt = "Add env spec %s to %s?" % (importable_spec.name, os.path.basename(project_file.filename))
            else:
                text = "Environment spec '%s' from %s is out of sync with %s. Diff:\n%s" % (
                    importable_spec.name, importable_filename, os.path.basename(project_file.filename),
                    importable_spec.diff_from(old))
                prompt = "Overwrite env spec %s with the changes from %s?" % (importable_spec.name, importable_filename)

            def overwrite_env_spec_from_importable(project):
                project.project_file.set_value(['env_specs', importable_spec.name], importable_spec.to_json())

            def remember_no_import_importable(project):
                project.project_file.set_value(['skip_imports', 'environment'], importable_spec.logical_hash)

            # we don't set the filename here because it isn't really an error in the
            # file, it ends up reading strangely.
            problems.append(
                ProjectProblem(
                    text=text,
                    fix_prompt=prompt,
                    fix_function=overwrite_env_spec_from_importable,
                    no_fix_function=remember_no_import_importable))
        elif env_specs_is_empty or env_specs_is_missing:
            # we do NOT want to add this problem if we merely
            # failed to parse individual env specs; it must be
            # safe to overwrite the env_specs key, so it has to
            # be empty or missing entirely. Also, we do NOT want
            # to add this if we are going to ask about environment.yml
            # import, above.
            def add_default_env_spec(project):
                default_spec = _anaconda_default_env_spec(self.global_base_env_spec)
                project.project_file.set_value(['env_specs', default_spec.name], default_spec.to_json())

            # This section should now be inaccessible
            # since env_spec will be added at runtime if missing


# this is only used for commands that don't specify anything
# (when/if we require all commands to specify, then remove this.)

        if 'default' in self.env_specs:
            self.default_env_spec_name = 'default'
        else:
            self.default_env_spec_name = first_env_spec_name

    def _update_conda_env_requirements(self, requirements, problems, project_file):
        if _fatal_problem(problems):
            return

        if self.has_bootstrap_env_spec():
            requirement = CondaBootstrapEnvRequirement(registry=self.registry, env_specs=self.env_specs)
            self._add_requirement(requirements, self.global_base_env_spec, requirement)

        requirement = CondaEnvRequirement(registry=self.registry, env_specs=self.env_specs)
        self._add_requirement(requirements, self.global_base_env_spec, requirement)

    def _update_commands(self, problems, project_file, requirements):
        failed = False

        first_command_name = None
        commands = dict()
        commands_section = project_file.get_value('commands', None)

        plugins = plugins_api.get_plugins('command_run')
        all_known_command_attributes_extended = (all_known_command_attributes + tuple(plugins.keys()))

        if commands_section is not None and not is_dict(commands_section):
            _file_problem(
                problems, project_file,
                "'commands:' section should be a dictionary from command names to attributes, not %r" %
                (commands_section))
            failed = True
        elif commands_section is not None:
            for (name, attrs) in commands_section.items():
                if name.strip() == '':
                    _file_problem(problems, project_file,
                                  "Command variable name cannot be empty string, found: '{}' as name".format(name))
                    failed = True
                    continue
                if first_command_name is None:
                    first_command_name = name

                if not is_dict(attrs):
                    _file_problem(
                        problems, project_file,
                        "command name '%s' should be followed by a dictionary of attributes not %r" % (name, attrs))
                    failed = True
                    continue

                _unknown_field_suggestions(project_file, problems, attrs, all_known_command_attributes_extended)

                if 'description' in attrs and not is_string(attrs['description']):
                    _file_problem(problems, project_file,
                                  "'description' field of command {} must be a string".format(name))
                    failed = True

                if 'supports_http_options' in attrs and not isinstance(attrs['supports_http_options'], bool):
                    _file_problem(problems, project_file,
                                  ("'supports_http_options' field of command {} must be a boolean".format(name)))
                    failed = True

                if 'env_spec' in attrs:
                    if not is_string(attrs['env_spec']):
                        _file_problem(
                            problems, project_file,
                            "'env_spec' field of command {} must be a string (an environment spec name)".format(name))
                        failed = True
                    elif attrs['env_spec'] not in self.env_specs:
                        _file_problem(
                            problems, project_file,
                            "env_spec '%s' for command '%s' does not appear in the env_specs section" %
                            (attrs['env_spec'], name))
                        failed = True

                if 'registers_fusion_function' in attrs and not isinstance(attrs['registers_fusion_function'], bool):
                    _file_problem(problems, project_file,
                                  ("'registers_fusion_function' field of command {} must be a boolean".format(name)))
                    failed = True

                copied_attrs = deepcopy(attrs)

                if 'env_spec' not in copied_attrs:
                    copied_attrs['env_spec'] = self.default_env_spec_name

                command_types = []
                ProjectCommandClass = ProjectCommand
                for attr in ALL_COMMAND_TYPES + tuple(plugins.keys()):
                    if attr not in copied_attrs:
                        continue
                    else:
                        if attr in plugins:
                            ProjectCommandClass = plugins[attr]

                    # be sure we add this even if the command is broken, since it's
                    # confusing to say "does not have a command line in it" below
                    # if the issue is that the command line is broken.
                    command_types.append(attr)

                    if not is_string(copied_attrs[attr]):
                        _file_problem(
                            problems, project_file, "command '%s' attribute '%s' should be a string not '%r'" %
                            (name, attr, copied_attrs[attr]))
                        failed = True

                if len(command_types) == 0:
                    _file_problem(problems, project_file, "command '%s' does not have a command line in it" % (name))
                    failed = True

                if ('notebook' in copied_attrs or 'bokeh_app' in copied_attrs) and (len(command_types) > 1):
                    label = 'bokeh_app' if 'bokeh_app' in copied_attrs else 'notebook'
                    others = copy(command_types)
                    others.remove(label)
                    others = [("'%s'" % other) for other in others]
                    _file_problem(
                        problems, project_file, "command '%s' has multiple commands in it, '%s' can't go with %s" %
                        (name, label, ", ".join(others)))
                    failed = True

                # note that once one command fails, we don't add any more
                if not failed:
                    commands[name] = ProjectCommandClass(name=name, attributes=copied_attrs)

        self._verify_notebook_commands(commands, problems, requirements, project_file)

        if failed:
            self.commands = dict()
            self.default_command_name = None
        else:
            self.commands = commands

        if 'default' in self.commands:
            self.default_command_name = 'default'
        else:
            # 'default' is always mapped to the first-listed if none is named 'default'
            # note: this may be None
            self.default_command_name = first_command_name

    def _verify_notebook_commands(self, commands, problems, requirements, project_file):
        skipped_notebooks = project_file.get_value(['skip_imports', 'notebooks'])
        if skipped_notebooks is not None:
            if skipped_notebooks is True:
                # skip ALL notebooks forever
                return
            elif not is_list(skipped_notebooks):
                _file_problem(
                    problems, project_file, "'skip_imports: notebooks:' value should be a list, found {}".format(
                        repr(skipped_notebooks)))
                return
        else:
            skipped_notebooks = []

        recorder = _new_error_recorder(_null_frontend())
        flat_requirements = []
        for reqs in requirements.values():
            flat_requirements.extend(reqs)
        files = _list_relative_paths_for_unignored_project_files(
            self.directory_path, frontend=recorder, requirements=flat_requirements)
        if files is None:
            problems.extend(recorder.pop_errors())
            assert problems != []
            return

        # chop out hidden directories. The
        # main reason to ignore dot directories is that they
        # might contain packages or git cache data or other
        # such gunk, not because we really care about
        # ".foo.ipynb" per se.
        files = [f for f in files if not f[0] == '.']

        # always use unix file separator
        files = [f.replace("\\", "/") for f in files]

        # use a deterministic order because the first command is the default
        files = sorted(files)

        def need_to_import_notebook(relative_name):
            for command in commands.values():
                if command.notebook == relative_name:
                    return False

            if relative_name in skipped_notebooks:
                return False

            return True

        def make_add_notebook_func(relative_name, env_spec_name):
            def add_notebook(project):
                errors = []
                extras = notebook_analyzer.extras(os.path.join(self.directory_path, relative_name), errors)
                # TODO this is broken, need to refactor so fix functions can return
                # errors and probably also log progress indication.
                assert [] == errors
                assert extras is not None

                command_dict = {'notebook': relative_name, 'env_spec': env_spec_name}
                command_dict.update(extras)
                project.project_file.set_value(['commands', relative_name], command_dict)

            return add_notebook

        def make_no_add_notebook_func(relative_name):
            def no_add_notebook(project):
                skipped_notebooks = project.project_file.get_value(['skip_imports', 'notebooks'], default=[])
                skipped_notebooks.append(relative_name)
                project.project_file.set_value(['skip_imports', 'notebooks'], skipped_notebooks)

            return no_add_notebook

        need_to_import = []
        for relative_name in files:
            if relative_name.endswith('.ipynb'):
                if need_to_import_notebook(relative_name):
                    need_to_import.append(relative_name)

        # make tests deterministic
        need_to_import.sort()

        if len(need_to_import) == 1:
            relative_name = need_to_import[0]
            problem = ProjectProblem(
                text="No command runs notebook %s" % (relative_name),
                filename=project_file.filename,
                fix_prompt="Create a command in %s for %s?" % (os.path.basename(project_file.filename), relative_name),
                fix_function=make_add_notebook_func(relative_name, self.default_env_spec_name),
                no_fix_function=make_no_add_notebook_func(relative_name),
                only_a_suggestion=True)
            problems.append(problem)
        elif len(need_to_import) > 1:
            add_funcs = [
                make_add_notebook_func(relative_name, self.default_env_spec_name) for relative_name in need_to_import
            ]
            no_add_funcs = [make_no_add_notebook_func(relative_name) for relative_name in need_to_import]

            def add_all(project):
                for f in add_funcs:
                    f(project)

            def no_add_all(project):
                for f in no_add_funcs:
                    f(project)

            problem = ProjectProblem(
                text="No commands run notebooks %s" % (", ".join(need_to_import)),
                filename=project_file.filename,
                fix_prompt="Create commands in %s for all missing notebooks?" % (os.path.basename(
                    project_file.filename)),
                fix_function=add_all,
                no_fix_function=no_add_all,
                only_a_suggestion=True)
            problems.append(problem)

    def _verify_command_dependencies(self, problems, project_file):
        for command in self.commands.values():
            if command.default_env_spec_name not in self.env_specs:
                # The missing environment will already have been flagged as a problem
                # in _update_commands, so we can just return
                return
            env_spec = self.env_specs[command.default_env_spec_name]
            missing = command.missing_packages(env_spec)
            if len(missing) > 0:

                def add_packages_to_env_spec(project):
                    env_dict = project.project_file.get_value(['env_specs', env_spec.name])
                    assert env_dict is not None
                    packages = env_dict.get('packages', [])
                    for m in missing:
                        # m would already be in there if we fixed the same env spec
                        # twice because two commands used it, for example.
                        if m not in packages:
                            packages.append(m)
                    project.project_file.set_value(['env_specs', env_spec.name, 'packages'], packages)

                problem = ProjectProblem(
                    text=("Command %s uses env spec %s which does not have the packages: %s" %
                          (command.name, env_spec.name, ", ".join(missing))),
                    filename=project_file.filename,
                    fix_prompt=("Add %s to env spec %s in %s?" % (", ".join(missing), env_spec.name,
                                                                  os.path.basename(project_file.filename))),
                    fix_function=add_packages_to_env_spec,
                    only_a_suggestion=True)
                problems.append(problem)

    def has_bootstrap_env_spec(self):
        """Return True if bootstrap-env is in env_specs, False otherwise."""
        return 'bootstrap-env' in self.env_specs


class Project(object):
    """Represents the information we've inferred about a project.

    The Project class encapsulates information from the project
    file, and also anything else we've guessed by snooping around in
    the project directory or global user configuration.
    """

    def __init__(self, directory_path, plugin_registry=None, frontend=None, must_exist=False):
        """Construct a Project with the given directory and plugin registry.

        Args:
            directory_path (str): path to the project directory
            plugin_registry (RequirementsRegistry): where to look up Requirement and Provider instances,
                                                    None for default
            frontend (Frontend): the UX using this Project instance
            must_exist (bool): if True, the absence of a project file is a problem
        """
        self._directory_path = os.path.realpath(directory_path).rstrip(os.sep)

        def load_default_specs():
            (importable_spec, importable_filename) = _find_importable_spec(directory_path)
            if importable_spec is not None:
                return [importable_spec]
            else:
                return [_anaconda_default_env_spec(shared_base_spec=None)]

        self._project_file = ProjectFile.load_for_directory(directory_path, default_env_specs_func=load_default_specs)
        self._lock_file = ProjectLockFile.load_for_directory(directory_path)
        self._directory_basename = os.path.basename(self._directory_path)
        self._config_cache = _ConfigCache(self._directory_path, plugin_registry, must_exist)
        if frontend is None:
            frontend = _null_frontend()
        assert isinstance(frontend, Frontend)
        self._frontends = [frontend]

    def _updated_cache(self):
        self._config_cache.update(self._project_file, self._lock_file)
        return self._config_cache

    @property
    def directory_path(self):
        """Get path to the project directory."""
        return self._directory_path

    @property
    def frontend(self):
        """Return the current UX frontend."""
        return self._frontends[-1]

    @property
    def project_file(self):
        """Get the ``ProjectFile`` for this project."""
        return self._project_file

    @property
    def lock_file(self):
        """Get the ``ProjectLockFile`` for this project."""
        return self._lock_file

    @property
    def plugin_registry(self):
        """Get the ``RequirementsRegistry`` for this project."""
        return self._config_cache.registry

    @property
    def union_of_requirements_for_all_envs(self):
        """Required items for ALL envs (list of ``Requirement`` instances)."""
        combined = []
        for env_spec in self.env_specs.values():
            # This is obviously pointless for the time being since
            # all env specs have the same requirements, but we want
            # to change that in the future so simulating it here.
            combined.extend(self.requirements(env_spec.name))
        return combined

    def requirements(self, env_spec_name):
        """Required items in order to run this project (list of ``Requirement`` instances)."""
        if env_spec_name is None:
            env_spec_name = self.default_env_spec_name
        requirements = self._updated_cache().requirements

        def get_reqs(env_spec):
            return requirements.get(env_spec.name, [])

        def req_key(req):
            assert isinstance(req, EnvVarRequirement)
            return req.env_var

        env_spec = self.env_specs.get(env_spec_name)

        if env_spec is None:
            # this happens if there was a problem parsing the project
            return []
        else:
            # this should probably really return the tuple instead
            # of list, but we'd have to fix up tests accordingly
            return list(env_spec._get_inherited_with_getter(get_reqs, key_func=req_key))

    def service_requirements(self, env_spec_name):
        """All requirements that are ServiceRequirement instances."""
        return self.find_requirements(env_spec_name, klass=ServiceRequirement)

    def download_requirements(self, env_spec_name):
        """All requirements that are DownloadRequirement instances."""
        return self.find_requirements(env_spec_name, klass=DownloadRequirement)

    def all_variable_requirements(self, env_spec_name):
        """All requirements that have an associated environment variable.

        Note: this will include services, downloads, and even CondaEnvRequirement.
        """
        return self.find_requirements(env_spec_name, klass=EnvVarRequirement)

    def plain_variable_requirements(self, env_spec_name):
        """All 'plain' variables (that aren't services, downloads, or a Conda environment for example).

        Use the ``all_variable_requirements`` property to get every variable.
        """
        return [req for req in self.all_variable_requirements(env_spec_name) if req.__class__ is EnvVarRequirement]

    def push_null_frontend(self):
        """Push a no-op frontend overriding the currently-active one.

        This is used to disable output temporarily.
        """
        self._frontends.append(_null_frontend())

    def pop_null_frontend(self):
        """Pop the no-op frontend."""
        assert len(self._frontends) > 1
        self._frontends = self._frontends[:-1]

    @contextlib.contextmanager
    def null_frontend(self):
        """Create a context with the frontend disabled."""
        self.push_null_frontend()
        try:
            yield
        finally:
            self.pop_null_frontend()

    def find_requirements(self, env_spec_name, env_var=None, klass=None):
        """Find requirements that match the given env var and class.

        If env_var and klass are both provided, BOTH must match.

        Args:
           env_spec_name (str): name of env spec to find requirements of
           env_var (str): if not None, filter requirements that have this env_var
           klass (class): if not None, filter requirements that are an instance of this class

        Returns:
           list of matching requirements (may be empty)
        """
        found = []
        for req in self.requirements(env_spec_name):
            if env_var is not None and not (isinstance(req, EnvVarRequirement) and req.env_var == env_var):
                continue
            if klass is not None and not isinstance(req, klass):
                continue
            found.append(req)
        return found

    @property
    def problems(self):
        """List of strings describing problems with the project configuration.

        This list contains problems which keep the project from loading, such as corrupt
        config files; it does not contain missing requirements and other "expected"
        problems.
        """
        return self._updated_cache().problem_strings

    @property
    def problem_objects(self):
        """List of ProjectProblem instances describing problems with the project configuration."""
        return [problem for problem in self._updated_cache().problems if not problem.only_a_suggestion]

    @property
    def fixable_problems(self):
        """List of ProjectProblem that have associated fix prompts."""
        return [p for p in self.problem_objects if p.can_fix and not p.only_a_suggestion]

    @property
    def unfixable_problems(self):
        """List of ProjectProblem that cannot be fixed."""
        return [p for p in self.problem_objects if not p.can_fix and not p.only_a_suggestion]

    def problems_status(self, description=None):
        """Get a ``Status`` describing project problems, or ``None`` if no problems."""
        if len(self.problems) > 0:
            errors = []
            for problem in self.problems:
                errors.append(problem)
            if description is None:
                description = "Unable to load the project."
            return SimpleStatus(success=False, description=description, errors=errors)
        else:
            return None

    @property
    def suggestions(self):
        """List of strings describing suggested changes to the project configuration."""
        return [problem.text for problem in self.suggestion_objects]

    @property
    def suggestion_objects(self):
        """List of ProjectProblem instances describing suggested changes to the project configuration."""
        return [problem for problem in self._updated_cache().problems if problem.only_a_suggestion]

    def fix_problems_and_suggestions(self):
        """Fix fixable problems and suggestions."""
        # the idea of this loop is that by fixing a problem we may
        # create a new one, for example we add a notebook command
        # and then the env spec needs to depend on "notebook".
        # However, we have no real way to detect an infinite
        # ping-pong of mutually-causing problems, so we cap
        # the iterations at an arbitrary number.
        iterations = 5
        while iterations > 0:
            fixed_a_thing = False
            for problem in self._updated_cache().problems:
                if problem.can_fix:
                    problem.fix(self)
                    fixed_a_thing = True
            if fixed_a_thing:
                self.use_changes_without_saving()
            iterations -= 1

    @property
    def name(self):
        """Get the project's human-readable name.

        Prefers in order: `name` field from anaconda-project.yml, `package:
        name:` from meta.yaml, then project directory name.
        """
        return self._updated_cache().name

    @property
    def url_friendly_name(self):
        """Get the project's url-friendly name."""
        return slugify(self.name)

    @property
    def description(self):
        """Get the project description."""
        return self._updated_cache().description

    @property
    def icon(self):
        """Get the project's icon as an absolute path or None if no icon.

        Prefers in order: `icon` field from anaconda-project.yml, `app:
        icon:` from meta.yaml.
        """
        return self._updated_cache().icon

    @property
    def env_specs(self):
        """Get a dictionary of environment names to CondaEnvironment instances."""
        return self._updated_cache().env_specs

    @property
    def locking_globally_enabled(self):
        """Get whether locking is enabled by default for lock sets that don't specify."""
        return self._updated_cache().locking_globally_enabled

    @property
    def global_base_env_spec(self):
        """Get the env spec representing global packages, channels, and platforms sections.

        This env spec has no name (its name is None) and can't be used directly
        to create environments, but every other env spec inherits from it.
        """
        return self._updated_cache().global_base_env_spec

    def all_variables(self, env_spec_name):
        """Get a list of strings with the variables names from ``all_variable_requirements``."""
        return [r.env_var for r in self.all_variable_requirements(env_spec_name)]

    def plain_variables(self, env_spec_name):
        """Get a list of strings with the variables names from ``plain_variable_requirements``."""
        return [r.env_var for r in self.plain_variable_requirements(env_spec_name)]

    def services(self, env_spec_name):
        """Get a list of strings with the variable names for the project services requirements."""
        return [r.env_var for r in self.service_requirements(env_spec_name)]

    def downloads(self, env_spec_name):
        """Get a list of strings with the variable names for the project download requirements."""
        return [r.env_var for r in self.download_requirements(env_spec_name)]

    @property
    def default_env_spec_name(self):
        """Get the named environment to use by default.

        This will be the one named "default" if it exists, and
        otherwise the first-listed one.

        Note that each command may have its own default, so
        this should only be used in contexts with no known
        command.
        """
        return self._updated_cache().default_env_spec_name

    def default_env_spec_name_for_command(self, command):
        """Get the named environment to use by default for a given ProjectCommand.

        the command may be ``None``
        """
        if command is None:
            return self.default_env_spec_name
        else:
            assert isinstance(command, ProjectCommand)
            return command.default_env_spec_name

    @property
    def commands(self):
        """Get the dictionary of commands to run the project.

        This dictionary can be empty.

        Returns:
            dictionary of command names to ``ProjectCommand``
        """
        return self._updated_cache().commands

    @property
    def default_command(self):
        """Get the default ``ProjectCommand`` or None if we don't have one.

        Returns:
            the default ``ProjectCommand``
        """
        cache = self._updated_cache()
        if cache.default_command_name is None:
            return None
        else:
            return cache.commands[cache.default_command_name]

    def default_exec_info_for_environment(self, environ, extra_args=None):
        """Get the information needed to run the project's default command.

        Args:
            environ (dict): the environment
            extra_args (list of str): extra args to append to the command line
        Returns:
            a CommandExecInfo instance
        """
        command = self.default_command
        if command is None:
            return None
        else:
            return command.exec_info_for_environment(environ=environ, extra_args=extra_args)

    def command_for_name(self, command_name):
        """Get the ProjectCommand for the given command name, or None if no commands.

        Args:
           command_name (str): the command name
        Returns:
           a ProjectCommand instance or None
        """
        if command_name is None:
            command_name = self._updated_cache().default_command_name
        if command_name is None:
            return None
        if command_name in self._updated_cache().commands:
            return self._updated_cache().commands[command_name]
        else:
            return None

    def publication_info(self):
        """Get JSON-serializable information to be stored as metadata when publishing the project.

        This is a "baked" version of anaconda-project.yml which also
        includes any defaults or automatic configuration.

        Before calling this, check that Project.problems is empty.

        Returns:
            A dictionary containing JSON-compatible types.
        """
        json = dict()
        json['name'] = self.name
        # the recipient will have to validate this; including it here
        # mostly because we might eventually allow the anaconda-project.yml to
        # manually provide it.
        json['url_friendly_name'] = self.url_friendly_name
        json['description'] = self.description
        json['anaconda_project_version'] = version
        commands = dict()
        for key, command in self.commands.items():
            commands[key] = dict(description=command.description)
            if command.bokeh_app is not None:
                commands[key]['bokeh_app'] = command.bokeh_app
            if command.notebook is not None:
                commands[key]['notebook'] = command.notebook
            if command.windows_cmd_commandline is not None:
                commands[key]['windows'] = command.windows_cmd_commandline
            if command.unix_shell_commandline is not None:
                commands[key]['unix'] = command.unix_shell_commandline
            if command is self.default_command:
                commands[key]['default'] = True
            commands[key]['env_spec'] = command.default_env_spec_name
            commands[key]['supports_http_options'] = command.supports_http_options
            commands[key].update(command.extras)
        json['commands'] = commands
        envs = dict()
        for key, env in self.env_specs.items():
            envs[key] = dict(
                packages=list(env.conda_packages),
                channels=list(env.channels),
                description=env.description,
                locked=env.lock_set.enabled,
                platforms=list(env.platforms))

            variables = dict()
            downloads = dict()
            services = dict()
            for req in self.requirements(key):
                if isinstance(req, CondaEnvRequirement):
                    continue

                if isinstance(req, EnvVarRequirement):
                    data = dict(title=req.title, description=req.description, encrypted=req.encrypted)

                    default = req.default_as_string
                    if default is not None:
                        data['default'] = default

                    if isinstance(req, DownloadRequirement):
                        data['url'] = req.url
                        downloads[req.env_var] = data
                    elif isinstance(req, ServiceRequirement):
                        data['type'] = req.service_type
                        services[req.env_var] = data
                    elif isinstance(req, EnvVarRequirement):
                        variables[req.env_var] = data

            envs[key]['downloads'] = downloads
            envs[key]['variables'] = variables
            envs[key]['services'] = services

        json['env_specs'] = envs

        return json

    def load(self):
        """Revert the project configuration by reloading config from disk.

        Discards all unsaved changes.

        This isn't needed when creating a Project, just if you want to
        revert to disk. We automatically load on Project creation.
        """
        self.project_file.load()
        self.lock_file.load()

    def save(self):
        """Save any modified project configuration.

        Does nothing for config files that are not dirty.
        """
        self.project_file.save()
        self.lock_file.save()

    def use_changes_without_saving(self):
        """Rebuild project state from in-memory changes.

        This causes the Project instance to reload from
        the in-memory (but possibly unsaved) state of
        the project file and lock file.
        """
        self.project_file.use_changes_without_saving()
        self.lock_file.use_changes_without_saving()

    @property
    def bootstrap_env_prefix(self):
        """Fullpath to bootstrap environment prefix."""
        return join(self._directory_path, 'envs', 'bootstrap-env')

    def is_running_in_bootstrap_env(self):
        """Return True if anaconda-project is running inside a project bootstrap env False otherwise."""
        return os.environ['CONDA_PREFIX'] == self.bootstrap_env_prefix

    def has_bootstrap_env_spec(self):
        """Return True if bootstrap-env is in env_specs, False otherwise."""
        return self._config_cache.has_bootstrap_env_spec()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2019, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""The ``download`` command retrieves a project archive from anaconda.org."""
from __future__ import absolute_import, print_function

from anaconda_project.internal.cli import console_utils
import anaconda_project.project_ops as project_ops


def download_command(
        project,
        unpack,
        parent_dir,
        site,
        username,
        token,
):
    """Download project from Anaconda Cloud.

    Returns:
        exit code
    """
    status = project_ops.download(
        project, unpack=unpack, parent_dir=parent_dir, site=site, username=username, token=token)
    if status:
        print(status.status_description)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def main(args):
    """Start the upload command and return exit status code."""
    return download_command(args.project, not args.no_unpack, args.parent_dir, args.site, args.user, args.token)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import codecs
import os

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.test.project_utils import assert_identical_except_blank_lines
from anaconda_project.project_file import ProjectFile, DEFAULT_PROJECT_FILENAME, possible_project_file_names
from anaconda_project.env_spec import EnvSpec

expected_default_file_template = """# This is an Anaconda project file.
#
# Here you can describe your project and how to run it.
# Use `anaconda-project run` to run the project.
# The file is in YAML format, please see http://www.yaml.org/start.html for more.
#

#
# Set the 'name' key to name your project
#
name: <NAME>
#
# Set the 'icon' key to give your project an icon
#
icon:
#
# Set a one-sentence-or-so 'description' key with project details
#
description:
#
# In the commands section, list your runnable scripts, notebooks, and other code.
# Use `anaconda-project add-command` to add commands.
#
commands: {}
#
# In the variables section, list any environment variables your code depends on.
# Use `anaconda-project add-variable` to add variables.
#
variables: {}
#
# In the services section, list any services that should be
# available before your code runs.
# Use `anaconda-project add-service` to add services.
#
services: {}
#
# In the downloads section, list any URLs to download to local files
# before your code runs.
# Use `anaconda-project add-download` to add downloads.
#
downloads: {}
%s%s%s%s"""

empty_global_packages = """#
# In the packages section, list any packages that must be installed
# before your code runs.
# Use `anaconda-project add-packages` to add packages.
#
packages: []
"""

empty_global_channels = """#
# In the channels section, list any Conda channel URLs to be searched
# for packages.
#
# For example,
#
# channels:
#    - mychannel
#
channels: []
"""

default_global_platforms = """#
# In the platforms section, list platforms the project should work on
# Examples: "linux-64", "osx-64", "win-64"
# Use `anaconda-project add-platforms` to add platforms.
#
platforms:
- linux-64
- osx-64
- win-64
"""

empty_default_env_specs = """#
# You can define multiple, named environment specs.
# Each inherits any global packages or channels,
# but can have its own unique ones also.
# Use `anaconda-project add-env-spec` to add environment specs.
#
env_specs:
  default:
    packages: []
    channels: []
"""


def _make_file_contents(packages, channels, platforms, env_specs):
    return expected_default_file_template % (packages, channels, platforms, env_specs)


expected_default_file = _make_file_contents(
    packages=empty_global_packages,
    channels=empty_global_channels,
    platforms=default_global_platforms,
    env_specs=empty_default_env_specs)


def test_create_missing_project_file():
    def create_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname)
        assert project_file is not None
        assert not os.path.exists(filename)
        project_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            expected = expected_default_file.replace("<NAME>", os.path.basename(dirname))
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)


def _use_existing_project_file(relative_name):
    def check_file(dirname):
        filename = os.path.join(dirname, relative_name)
        assert os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname)
        value = project_file.get_value(["a", "b"])
        assert "c" == value

    with_directory_contents({relative_name: "a:\n  b: c"}, check_file)


def test_use_existing_project_file_default_name():
    _use_existing_project_file(DEFAULT_PROJECT_FILENAME)


def test_use_existing_project_file_all_names():
    for name in possible_project_file_names:
        _use_existing_project_file(name)


def test_load_directory_without_project_file():
    def read_missing_file(dirname):
        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname)
        assert project_file is not None
        assert not os.path.exists(filename)
        assert project_file.get_value(["a", "b"]) is None

    with_directory_contents(dict(), read_missing_file)


anaconda_global_packages = """#
# In the packages section, list any packages that must be installed
# before your code runs.
# Use `anaconda-project add-packages` to add packages.
#
packages:
- anaconda
"""

mychannel_global_channels = """#
# In the channels section, list any Conda channel URLs to be searched
# for packages.
#
# For example,
#
# channels:
#    - mychannel
#
channels:
- mychannel
"""

abc_empty_env_spec = """#
# You can define multiple, named environment specs.
# Each inherits any global packages or channels,
# but can have its own unique ones also.
# Use `anaconda-project add-env-spec` to add environment specs.
#
env_specs:
  abc:
    description: ABC
    packages: []
    channels: []
"""

expected_one_env_spec_contents = _make_file_contents(
    packages=anaconda_global_packages,
    channels=mychannel_global_channels,
    platforms=default_global_platforms,
    env_specs=abc_empty_env_spec)


def test_create_missing_project_file_one_default_env_spec():
    def create_file(dirname):
        def default_env_specs_func():
            return [
                EnvSpec(
                    name='abc',
                    conda_packages=['anaconda'],
                    pip_packages=[],
                    channels=['mychannel'],
                    description="ABC",
                    inherit_from_names=(),
                    inherit_from=())
            ]

        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname, default_env_specs_func=default_env_specs_func)
        assert project_file is not None
        assert not os.path.exists(filename)
        project_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            expected = expected_one_env_spec_contents.replace("<NAME>", os.path.basename(dirname))
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)


def test_create_missing_project_file_one_default_env_spec_deps():
    def create_file(dirname):
        def default_env_specs_func():
            return [
                EnvSpec(
                    name='abc',
                    conda_packages=['anaconda'],
                    pip_packages=[],
                    channels=['mychannel'],
                    description="ABC",
                    inherit_from_names=(),
                    inherit_from=())
            ]

        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname, default_env_specs_func=default_env_specs_func)
        assert project_file is not None
        assert not os.path.exists(filename)
        project_file.save()
        assert os.path.exists(filename)
        print(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            contents = contents.replace('packages', 'dependencies')
            expected_one_env_spec_contents_deps = expected_one_env_spec_contents.replace('packages', 'dependencies')
            expected = expected_one_env_spec_contents_deps.replace("<NAME>", os.path.basename(dirname))
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)


abc_xyz_env_specs = """#
# You can define multiple, named environment specs.
# Each inherits any global packages or channels,
# but can have its own unique ones also.
# Use `anaconda-project add-env-spec` to add environment specs.
#
env_specs:
  abc:
    description: ABC
    packages:
    - anaconda
    channels:
    - mychannel
  xyz:
    description: XYZ
    packages:
    - foo
    channels:
    - bar
"""

expected_two_env_spec_contents = _make_file_contents(
    packages=empty_global_packages,
    channels=empty_global_channels,
    platforms=default_global_platforms,
    env_specs=abc_xyz_env_specs)


def test_create_missing_project_file_two_default_env_specs():
    def create_file(dirname):
        def default_env_specs_func():
            return [
                EnvSpec(
                    name='abc',
                    conda_packages=['anaconda'],
                    pip_packages=[],
                    channels=['mychannel'],
                    description="ABC",
                    inherit_from_names=(),
                    inherit_from=()),
                EnvSpec(
                    name='xyz',
                    conda_packages=['foo'],
                    pip_packages=[],
                    channels=['bar'],
                    description="XYZ",
                    inherit_from_names=(),
                    inherit_from=())
            ]

        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname, default_env_specs_func=default_env_specs_func)
        assert project_file is not None
        assert not os.path.exists(filename)
        project_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            expected = expected_two_env_spec_contents.replace("<NAME>", os.path.basename(dirname))
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)


def test_create_missing_project_file_two_default_env_specs_deps():
    def create_file(dirname):
        def default_env_specs_func():
            return [
                EnvSpec(
                    name='abc',
                    conda_packages=['anaconda'],
                    pip_packages=[],
                    channels=['mychannel'],
                    description="ABC",
                    inherit_from_names=(),
                    inherit_from=()),
                EnvSpec(
                    name='xyz',
                    conda_packages=['foo'],
                    pip_packages=[],
                    channels=['bar'],
                    description="XYZ",
                    inherit_from_names=(),
                    inherit_from=())
            ]

        filename = os.path.join(dirname, DEFAULT_PROJECT_FILENAME)
        assert not os.path.exists(filename)
        project_file = ProjectFile.load_for_directory(dirname, default_env_specs_func=default_env_specs_func)
        assert project_file is not None
        assert not os.path.exists(filename)
        project_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            contents = contents.replace('packages', 'dependencies')
            expected_two_env_spec_contents_deps = expected_two_env_spec_contents.replace('packages', 'dependencies')
            expected = expected_two_env_spec_contents_deps.replace("<NAME>", os.path.basename(dirname))
            assert_identical_except_blank_lines(expected, contents)

    with_directory_contents(dict(), create_file)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from anaconda_project.yaml_file import YamlFile
from anaconda_project.internal.test.tmpfile_utils import with_file_contents, with_directory_contents

import errno
import os
import pytest


def test_read_yaml_file_and_get_value():
    def check_abc(filename):
        yaml = YamlFile(filename)
        assert not yaml.corrupted
        assert yaml.corrupted_error_message is None
        assert yaml.change_count == 1
        # try getting with a list of keys
        value = yaml.get_value(["a", "b"])
        assert "c" == value
        # get a single string as the path
        value = yaml.get_value("a")
        assert dict(b="c") == value
        # get with a tuple to show we aren't list-specific
        value = yaml.get_value(("a", "b"))
        assert "c" == value

        assert yaml.root == dict(a=dict(b='c'))

    with_file_contents("""
a:
  b: c
""", check_abc)


def test_read_yaml_file_and_get_default():
    def check_abc(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "z"], "default")
        assert "default" == value

    with_file_contents("""
a:
  b: c
""", check_abc)


def test_read_empty_yaml_file_and_get_default():
    def check_empty(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "z"], "default")
        assert "default" == value

    with_file_contents("", check_empty)


def test_read_yaml_file_and_get_list_valued_section():
    def get_list_value(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value("a")
        assert [1, 2, 3] == value

    with_file_contents("""
a: [1,2,3]
""", get_list_value)


def test_read_yaml_file_and_get_default_due_to_missing_section():
    def check_abc(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value(["z", "b"], "default")
        assert "default" == value

    with_file_contents("""
a:
  b: c
""", check_abc)


def test_read_yaml_file_and_get_default_due_to_non_dict_section():
    def check_a(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "b"], "default")
        assert "default" == value

    with_file_contents("""
a: 42
""", check_a)


def test_invalid_path():
    def check_bad_path(filename):
        yaml = YamlFile(filename)
        assert not yaml.corrupted
        with pytest.raises(ValueError) as excinfo:
            yaml.get_value(42)
        assert "YAML file path must be a string or an iterable of strings" in repr(excinfo.value)

    with_file_contents("""
a:
  b: c
""", check_bad_path)


def test_read_missing_yaml_file_and_get_default_due_to_missing_section():
    def check_missing(dirname):
        yaml = YamlFile(os.path.join(dirname, "nope.yaml"))
        value = yaml.get_value(["z", "b"], "default")
        assert "default" == value

    with_directory_contents(dict(), check_missing)


def test_read_empty_yaml_file_and_get_default_due_to_missing_section():
    def check_missing(dirname):
        yaml = YamlFile(os.path.join(dirname, "nope.yaml"))
        value = yaml.get_value(["z", "b"], "default")
        assert "default" == value

    with_directory_contents({"nope.yaml": ""}, check_missing)


def test_read_yaml_file_that_is_a_directory():
    def check_read_directory(dirname):
        filename = os.path.join(dirname, "dir.yaml")
        os.makedirs(filename)
        with pytest.raises(IOError) as excinfo:
            YamlFile(filename)
        import platform
        if platform.system() == 'Windows':
            assert errno.EACCES == excinfo.value.errno
        else:
            assert errno.EISDIR == excinfo.value.errno

    with_directory_contents(dict(), check_read_directory)


def test_read_yaml_file_and_change_value():
    # ruamel.yaml does reformat yaml files a little bit,
    # for example it picks its own indentation, even
    # as it tries to keep comments and stuff. So
    # this test cheats by using input that happens
    # to be in the format ruamel.yaml will generate.
    # Oh well.
    template = """
# this is a comment 1
a:
  # this is a comment 2
  b: %s
"""

    template = template[1:]  # chop leading newline

    original_value = "c"
    original_content = template % (original_value)
    changed_value = 42
    changed_content = template % (changed_value)

    def change_abc(filename):
        yaml = YamlFile(filename)
        assert yaml.change_count == 1
        value = yaml.get_value(["a", "b"])
        assert original_value == value
        yaml.set_value(["a", "b"], changed_value)
        yaml.save()

        import codecs
        with codecs.open(filename, 'r', 'utf-8') as file:
            changed = file.read()
            assert changed_content == changed

        yaml2 = YamlFile(filename)
        assert yaml2.change_count == 1
        value2 = yaml2.get_value(["a", "b"])
        assert changed_value == value2

    with_file_contents(original_content, change_abc)


def test_read_missing_yaml_file_and_set_value():
    def set_abc(dirname):
        filename = os.path.join(dirname, "foo.yaml")
        assert not os.path.exists(filename)
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "b"])
        assert value is None
        yaml.set_value(["a", "b"], 42)
        yaml.save()
        assert os.path.exists(filename)

        import codecs
        with codecs.open(filename, 'r', 'utf-8') as file:
            changed = file.read()
            expected = """
# yaml file
a:
  b: 42
""" [1:]

            assert expected == changed

        yaml2 = YamlFile(filename)
        value2 = yaml2.get_value(["a", "b"])
        assert 42 == value2

    with_directory_contents(dict(), set_abc)


def test_read_empty_yaml_file_and_set_value():
    def set_abc(dirname):
        filename = os.path.join(dirname, "foo.yaml")
        assert os.path.exists(filename)
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "b"])
        assert value is None
        yaml.set_value(["a", "b"], 42)
        yaml.save()
        assert os.path.exists(filename)

        import codecs
        with codecs.open(filename, 'r', 'utf-8') as file:
            changed = file.read()
            expected = """
a:
  b: 42
""" [1:]

            assert expected == changed

        yaml2 = YamlFile(filename)
        value2 = yaml2.get_value(["a", "b"])
        assert 42 == value2

    with_directory_contents({"foo.yaml": ""}, set_abc)


def test_read_yaml_file_and_add_section():
    original_content = """
a:
  b: c
"""

    def add_section(filename):
        yaml = YamlFile(filename)
        value = yaml.get_value(["a", "b"])
        assert "c" == value
        yaml.set_value(["x", "y"], dict(z=42, q="rs"))
        assert yaml.change_count == 1
        yaml.save()
        assert yaml.change_count == 2

        yaml2 = YamlFile(filename)
        value2 = yaml2.get_value(["a", "b"])
        assert "c" == value2

        added_value = yaml2.get_value(["x", "y", "z"])
        assert 42 == added_value

        added_value_2 = yaml2.get_value(["x", "y", "q"])
        assert "rs" == added_value_2

        print(open(filename, 'r').read())

    with_file_contents(original_content, add_section)


def test_multiple_saves_ignored_if_not_dirty():
    def check_dirty_handling(dirname):
        filename = os.path.join(dirname, "foo.yaml")
        assert not os.path.exists(filename)
        yaml = YamlFile(filename)
        assert yaml.change_count == 1
        yaml.set_value(["a", "b"], 42)
        yaml.save()
        assert yaml.change_count == 2
        assert os.path.exists(filename)
        time1 = os.path.getmtime(filename)

        yaml.save()
        assert time1 == os.path.getmtime(filename)
        assert yaml.change_count == 2
        yaml.save()
        assert time1 == os.path.getmtime(filename)
        assert yaml.change_count == 2

        yaml.set_value(["a", "b"], 43)
        assert time1 == os.path.getmtime(filename)
        assert yaml.change_count == 2
        yaml.save()
        # OS mtime resolution might leave these equal
        assert time1 <= os.path.getmtime(filename)
        assert yaml.change_count == 3

    with_directory_contents(dict(), check_dirty_handling)


def test_save_ignored_if_not_dirty_after_load():
    def check_dirty_handling(dirname):
        filename = os.path.join(dirname, "foo.yaml")
        assert not os.path.exists(filename)
        yaml = YamlFile(filename)
        yaml.set_value(["a", "b"], 42)
        yaml.save()
        assert os.path.exists(filename)
        time1 = os.path.getmtime(filename)

        yaml2 = YamlFile(filename)
        assert time1 == os.path.getmtime(filename)
        assert yaml2.change_count == 1
        yaml2.save()
        assert time1 == os.path.getmtime(filename)
        assert yaml2.change_count == 1

    with_directory_contents(dict(), check_dirty_handling)


def test_throw_if_cannot_create_directory(monkeypatch):
    def mock_makedirs(path, mode=0):
        raise IOError("this is not EEXIST")

    monkeypatch.setattr("os.makedirs", mock_makedirs)

    def check_throw_if_cannot_create(dirname):
        subdir = "bar"
        filename = os.path.join(dirname, subdir, "foo.yaml")

        yaml = YamlFile(filename)
        yaml.set_value(["a", "b"], 42)
        with pytest.raises(IOError) as excinfo:
            yaml.save()
        assert "this is not EEXIST" in repr(excinfo.value)

    with_directory_contents(dict(), check_throw_if_cannot_create)


def test_read_corrupted_yaml_file():
    def check_corrupted(filename):
        yaml = YamlFile(filename)
        assert yaml.corrupted
        assert "mapping values are not allowed here" in yaml.corrupted_error_message

        # it should raise an exception if you try to modify
        with pytest.raises(ValueError) as excinfo:
            yaml.set_value(["foo", "bar"], 42)
        assert "Cannot modify corrupted" in repr(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            yaml.save()
        assert "Cannot modify corrupted" in repr(excinfo.value)

        # the file should appear empty if you try to get anything,
        # but it shouldn't throw
        assert yaml._yaml is not None
        assert yaml.get_value(["a", "b"]) is None

    with_file_contents("""
^
a:
  b: c
""", check_corrupted)


def test_roundtrip_yaml_file_preserving_order_and_comments():
    original_content = """
# comment in front of a
a:
  x: y
  # comment in front of z
  z: q

b:
  i: j

  # whitespace in front of this comment in front of k
  k: l

c:
  # comment before a list item
  - foo
  - bar # comment after a list item

d:
  hello: world
  foo: bar

e:
  woot: woot
  # comment at the end of e

# comment in column 0 at the end
# this one is a block comment
# which continues several lines


"""

    def check_roundtrip(filename):
        yaml = YamlFile(filename)
        yaml._previous_content = "not the actual previous content"
        yaml.save()
        new_content = open(filename, 'r').read()
        print("the re-saved version of the file was:")
        print(new_content)
        assert original_content != new_content

        # We don't require that the YAML backend preserves every
        # formatting detail, but it can't reorder things or lose
        # comments because if it did users would be annoyed.
        # Minor whitespace changes are OK, though ideally we'd
        # avoid even those.
        def canonicalize(content):
            if content.startswith("\n"):
                content = content[1:]
            return content.replace(" ", "").replace("\n\n", "\n")

        original_canon = canonicalize(original_content)
        new_canon = canonicalize(new_content)
        assert original_canon == new_canon

    with_file_contents(original_content, check_roundtrip)


def test_read_yaml_file_and_unset_values():
    # testing single-item dict, two-item dict, and toplevel value
    original_content = """
a:
  b: 1

x:
  y: 2
  z: 3

q: 4
"""

    def unset_values(filename):
        yaml = YamlFile(filename)
        assert yaml.change_count == 1
        a_b = yaml.get_value(["a", "b"])
        assert 1 == a_b
        x_y = yaml.get_value(["x", "y"])
        assert 2 == x_y
        x_z = yaml.get_value(["x", "z"])
        assert 3 == x_z
        q = yaml.get_value("q")
        assert 4 == q

        def assert_unset_on_reload(path):
            yaml2 = YamlFile(filename)
            assert yaml2.change_count == 1
            value2 = yaml2.get_value(path, None)
            assert value2 is None

        scope = dict(last_change=yaml.change_count)

        def check_unset(path):
            assert yaml.change_count == scope['last_change']
            assert not yaml.has_unsaved_changes
            yaml.unset_value(path)
            assert yaml.get_value(path, None) is None
            assert yaml.has_unsaved_changes
            yaml.save()
            assert yaml.change_count == (scope['last_change'] + 1)
            scope['last_change'] += 1
            assert_unset_on_reload(path)

        check_unset(["a", "b"])
        check_unset(["x", "y"])
        check_unset(["x", "z"])
        check_unset("q")

        assert not yaml.has_unsaved_changes
        yaml.unset_value("not_in_there")
        assert not yaml.has_unsaved_changes

    with_file_contents(original_content, unset_values)


def test_read_yaml_file_and_set_get_empty_string():
    def check(filename):
        yaml = YamlFile(filename)
        assert not yaml.corrupted
        assert yaml.corrupted_error_message is None
        assert yaml.change_count == 1
        value = yaml.get_value("a", None)
        assert value is None

        yaml.set_value("a", '')
        value = yaml.get_value("a", None)
        assert value == ''

        # only-whitespace string
        yaml.set_value("a", ' ')
        value = yaml.get_value("a", None)
        assert value == ' '

    with_file_contents("", check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""The ``main`` function chooses and runs a subcommand."""
from __future__ import absolute_import, print_function

import logging
import os
import sys
from argparse import ArgumentParser, REMAINDER

from anaconda_project.internal.cli.prepare_with_mode import (UI_MODE_TEXT_ASK_QUESTIONS,
                                                             UI_MODE_TEXT_DEVELOPMENT_DEFAULTS_OR_ASK, _all_ui_modes)
from anaconda_project.version import version
from anaconda_project.verbose import push_verbose_logger, pop_verbose_logger
from anaconda_project.project import ALL_COMMAND_TYPES
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirements.download import _hash_algorithms
import anaconda_project
from anaconda_project.internal.cli.bug_handler import handle_bugs
import anaconda_project.internal.cli.init as init
import anaconda_project.internal.cli.run as run
import anaconda_project.internal.cli.prepare as prepare
import anaconda_project.internal.cli.clean as clean
import anaconda_project.internal.cli.archive as archive
import anaconda_project.internal.cli.unarchive as unarchive
import anaconda_project.internal.cli.upload as upload
import anaconda_project.internal.cli.download as download
import anaconda_project.internal.cli.activate as activate
import anaconda_project.internal.cli.variable_commands as variable_commands
import anaconda_project.internal.cli.download_commands as download_commands
import anaconda_project.internal.cli.service_commands as service_commands
import anaconda_project.internal.cli.environment_commands as environment_commands
import anaconda_project.internal.cli.command_commands as command_commands


def _parse_args_and_run_subcommand(argv):
    parser = ArgumentParser(prog="anaconda-project", description="Actions on projects (runnable projects).")

    subparsers = parser.add_subparsers(help="Sub-commands")

    parser.add_argument('-v', '--version', action='version', version=version)
    parser.add_argument('--verbose', action='store_true', default=False, help="show verbose debugging details")

    def add_directory_arg(preset):
        preset.add_argument(
            '--directory',
            metavar='PROJECT_DIR',
            default='.',
            help="Project directory containing anaconda-project.yml (defaults to current directory)")

    def add_env_spec_arg(preset):
        preset.add_argument(
            '--env-spec',
            metavar='ENVIRONMENT_SPEC_NAME',
            default=None,
            action='store',
            help="An environment spec name from anaconda-project.yml")

    def add_prepare_args(preset, include_command=True):
        add_directory_arg(preset)
        add_env_spec_arg(preset)
        all_supported_modes = list(_all_ui_modes)
        # we don't support "ask about every single thing" mode yet.
        all_supported_modes.remove(UI_MODE_TEXT_ASK_QUESTIONS)
        preset.add_argument(
            '--mode',
            metavar='MODE',
            default=UI_MODE_TEXT_DEVELOPMENT_DEFAULTS_OR_ASK,
            choices=_all_ui_modes,
            action='store',
            help="One of " + ", ".join(_all_ui_modes))
        if include_command:
            preset.add_argument(
                '--command',
                metavar='COMMAND_NAME',
                default=None,
                action='store',
                help="A command name from anaconda-project.yml (env spec for this command will be used)")

    def add_env_spec_name_arg(preset, required):
        preset.add_argument(
            '-n',
            '--name',
            metavar='ENVIRONMENT_SPEC_NAME',
            required=required,
            action='store',
            help="Name of the environment spec from anaconda-project.yml")

    preset = subparsers.add_parser('init', help="Initialize a directory with default project configuration")
    add_directory_arg(preset)
    preset.add_argument(
        '--empty-environment',
        action='store_true',
        help="Do not add the default package set to the environment.",
        default=None)
    preset.add_argument('-y', '--yes', action='store_true', help="Assume yes to all confirmation prompts", default=None)
    preset.set_defaults(main=init.main)

    preset = subparsers.add_parser('run', help="Run the project, setting up requirements first")
    add_prepare_args(preset, include_command=False)
    preset.add_argument(
        'command', metavar='COMMAND_NAME', default=None, nargs='?', help="A command name from anaconda-project.yml")
    preset.add_argument('extra_args_for_command', metavar='EXTRA_ARGS_FOR_COMMAND', default=None, nargs=REMAINDER)
    preset.set_defaults(main=run.main)

    preset = subparsers.add_parser('prepare', help="Set up the project requirements, but does not run the project")
    preset.add_argument('--all', action='store_true', help="Prepare all environments", default=None)
    preset.add_argument('--refresh', action='store_true', help='Remove and recreate the environment', default=None)
    add_prepare_args(preset)
    preset.set_defaults(main=prepare.main)

    preset = subparsers.add_parser(
        'clean', help="Removes generated state (stops services, deletes environment files, etc)")
    add_directory_arg(preset)
    preset.set_defaults(main=clean.main)

    if not anaconda_project._beta_test_mode:
        preset = subparsers.add_parser(
            'activate', help="Set up the project and output shell export commands reflecting the setup")
        add_prepare_args(preset)
        preset.set_defaults(main=activate.main)

    preset = subparsers.add_parser(
        'archive', help="Create a .zip, .tar.gz, or .tar.bz2 archive with project files in it")
    add_directory_arg(preset)
    preset.add_argument('filename', metavar='ARCHIVE_FILENAME')
    preset.set_defaults(main=archive.main)

    preset = subparsers.add_parser(
        'unarchive', help="Unpack a .zip, .tar.gz, or .tar.bz2 archive with project files in it")
    preset.add_argument('filename', metavar='ARCHIVE_FILENAME')
    preset.add_argument('directory', metavar='DESTINATION_DIRECTORY', default=None, nargs='?')

    preset.set_defaults(main=unarchive.main)

    preset = subparsers.add_parser('upload', help="Upload the project to Anaconda Cloud")
    add_directory_arg(preset)
    preset.add_argument('-p', '--private', action='store_true', help="Upload a private project", default=None)
    preset.add_argument('-s', '--site', metavar='SITE', help='Select site to use')
    preset.add_argument('-t', '--token', metavar='TOKEN', help='Auth token or a path to a file containing a token')
    preset.add_argument('-u', '--user', metavar='USERNAME', help='User account, defaults to the current user')
    preset.add_argument(
        '--suffix',
        metavar='SUFFIX',
        help='Project archive suffix (.tar.gz, .tar.bz2, .zip)',
        default='.tar.bz2',
        choices=['.tar.gz', '.tar.bz2', '.zip'])
    preset.set_defaults(main=upload.main)

    preset = subparsers.add_parser('download', help="Download the project from Anaconda Cloud")
    add_directory_arg(preset)
    preset.add_argument(
        'project',
        help='The project to download as <username>/<projectname>. If <projectname>' +
        'has spaces inclose everything in quotes "<username>/<project name>".' +
        'If specified as <projectname> then the logged-in username is used.')
    preset.add_argument('--no-unpack', action='store_true', help='Do not unpack the project archive.')
    preset.add_argument(
        '--parent_dir',
        default=None,
        help='Download archive to specific directory, otherwise downloaded to current working directory.')
    preset.add_argument('-s', '--site', metavar='SITE', help='Select site to use')
    preset.add_argument('-t', '--token', metavar='TOKEN', help='Auth token or a path to a file containing a token')
    preset.add_argument('-u', '--user', metavar='USERNAME', help='User account, defaults to the current user')
    preset.set_defaults(main=download.main)

    preset = subparsers.add_parser('add-variable', help="Add a required environment variable to the project")
    add_env_spec_arg(preset)
    preset.add_argument('vars_to_add', metavar='VARS_TO_ADD', default=None, nargs=REMAINDER)
    preset.add_argument(
        '--default', metavar='DEFAULT_VALUE', default=None, help='Default value if environment variable is unset')
    add_directory_arg(preset)
    preset.set_defaults(main=variable_commands.main_add)

    preset = subparsers.add_parser('remove-variable', help="Remove an environment variable from the project")
    add_env_spec_arg(preset)
    add_directory_arg(preset)
    preset.add_argument('vars_to_remove', metavar='VARS_TO_REMOVE', default=None, nargs=REMAINDER)
    preset.set_defaults(main=variable_commands.main_remove)

    preset = subparsers.add_parser('list-variables', help="List all variables on the project")
    add_env_spec_arg(preset)
    add_directory_arg(preset)
    preset.set_defaults(main=variable_commands.main_list)

    preset = subparsers.add_parser(
        'set-variable', help="Set an environment variable value in anaconda-project-local.yml")
    add_env_spec_arg(preset)
    preset.add_argument('vars_and_values', metavar='VARS_AND_VALUES', default=None, nargs=REMAINDER)
    add_directory_arg(preset)
    preset.set_defaults(main=variable_commands.main_set)

    preset = subparsers.add_parser(
        'unset-variable', help="Unset an environment variable value from anaconda-project-local.yml")
    add_env_spec_arg(preset)
    add_directory_arg(preset)
    preset.add_argument('vars_to_unset', metavar='VARS_TO_UNSET', default=None, nargs=REMAINDER)
    preset.set_defaults(main=variable_commands.main_unset)

    preset = subparsers.add_parser('add-download', help="Add a URL to be downloaded before running commands")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.add_argument('filename_variable', metavar='ENV_VAR_FOR_FILENAME', default=None)
    preset.add_argument('download_url', metavar='DOWNLOAD_URL', default=None)
    preset.add_argument('--filename', help="The name to give the file/folder after downloading it", default=None)
    preset.add_argument(
        '--hash-algorithm', help="Defines which hash algorithm to use", default=None, choices=_hash_algorithms)
    preset.add_argument('--hash-value', help="The expected checksum hash of the downloaded file", default=None)
    preset.set_defaults(main=download_commands.main_add)

    preset = subparsers.add_parser('remove-download', help="Remove a download from the project and from the filesystem")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.add_argument('filename_variable', metavar='ENV_VAR_FOR_FILENAME', default=None)
    preset.set_defaults(main=download_commands.main_remove)

    preset = subparsers.add_parser('list-downloads', help="List all downloads on the project")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.set_defaults(main=download_commands.main_list)

    service_types = RequirementsRegistry().list_service_types()
    service_choices = list(map(lambda s: s.name, service_types))

    def add_service_variable_name(preset):
        preset.add_argument('--variable', metavar='ENV_VAR_FOR_SERVICE_ADDRESS', default=None)

    preset = subparsers.add_parser('add-service', help="Add a service to be available before running commands")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    add_service_variable_name(preset)
    preset.add_argument('service_type', metavar='SERVICE_TYPE', default=None, choices=service_choices)
    preset.set_defaults(main=service_commands.main_add)

    preset = subparsers.add_parser('remove-service', help="Remove a service from the project")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.add_argument('variable', metavar='SERVICE_REFERENCE', default=None)
    preset.set_defaults(main=service_commands.main_remove)

    preset = subparsers.add_parser('list-services', help="List services present in the project")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.set_defaults(main=service_commands.main_list)

    def add_package_args(preset):
        preset.add_argument(
            '-c', '--channel', metavar='CHANNEL', action='append', help='Channel to search for packages')
        preset.add_argument('packages', metavar='PACKAGES', default=None, nargs=REMAINDER)

    preset = subparsers.add_parser('add-env-spec', help="Add a new environment spec to the project")
    add_directory_arg(preset)
    add_package_args(preset)
    add_env_spec_name_arg(preset, required=True)
    preset.set_defaults(main=environment_commands.main_add)

    preset = subparsers.add_parser('remove-env-spec', help="Remove an environment spec from the project")
    add_directory_arg(preset)
    add_env_spec_name_arg(preset, required=True)
    preset.set_defaults(main=environment_commands.main_remove)

    preset = subparsers.add_parser('list-env-specs', help="List all environment specs for the project")
    add_directory_arg(preset)
    preset.set_defaults(main=environment_commands.main_list_env_specs)

    preset = subparsers.add_parser('export-env-spec', help="Save an environment spec as a conda environment file")
    add_directory_arg(preset)
    add_env_spec_name_arg(preset, required=False)
    preset.add_argument('filename', metavar='ENVIRONMENT_FILE')
    preset.set_defaults(main=environment_commands.main_export)

    preset = subparsers.add_parser('lock', help="Lock all packages at their current versions")
    add_directory_arg(preset)
    add_env_spec_name_arg(preset, required=False)
    preset.set_defaults(main=environment_commands.main_lock)

    preset = subparsers.add_parser('unlock', help="Remove locked package versions")
    add_directory_arg(preset)
    add_env_spec_name_arg(preset, required=False)
    preset.set_defaults(main=environment_commands.main_unlock)

    preset = subparsers.add_parser('update', help="Update all packages to their latest versions")
    add_directory_arg(preset)
    add_env_spec_name_arg(preset, required=False)
    preset.set_defaults(main=environment_commands.main_update)

    preset = subparsers.add_parser('add-packages', help="Add packages to one or all project environments")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    add_package_args(preset)
    preset.set_defaults(main=environment_commands.main_add_packages)

    preset = subparsers.add_parser('remove-packages', help="Remove packages from one or all project environments")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.add_argument('packages', metavar='PACKAGE_NAME', default=None, nargs='+')
    preset.set_defaults(main=environment_commands.main_remove_packages)

    preset = subparsers.add_parser('list-packages', help="List packages for an environment on the project")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.set_defaults(main=environment_commands.main_list_packages)

    def add_platforms_list(preset):
        preset.add_argument('platforms', metavar='PLATFORM_NAME', default=None, nargs='+')

    preset = subparsers.add_parser('add-platforms', help="Add platforms to one or all project environments")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    add_platforms_list(preset)
    preset.set_defaults(main=environment_commands.main_add_platforms)

    preset = subparsers.add_parser('remove-platforms', help="Remove platforms from one or all project environments")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    add_platforms_list(preset)
    preset.set_defaults(main=environment_commands.main_remove_platforms)

    preset = subparsers.add_parser('list-platforms', help="List platforms for an environment on the project")
    add_directory_arg(preset)
    add_env_spec_arg(preset)
    preset.set_defaults(main=environment_commands.main_list_platforms)

    def add_command_name_arg(preset):
        preset.add_argument('name', metavar="NAME", help="Command name used to invoke it")

    preset = subparsers.add_parser('add-command', help="Add a new command to the project")
    add_directory_arg(preset)
    command_choices = list(ALL_COMMAND_TYPES) + ['ask']
    command_choices.remove("conda_app_entry")  # conda_app_entry is sort of silly and may go away
    preset.add_argument('--type', action="store", choices=command_choices, help="Command type to add")
    add_command_name_arg(preset)
    add_env_spec_arg(preset)
    preset.add_argument(
        '--supports-http-options',
        dest='supports_http_options',
        action="store_true",
        help="The command supports project's HTTP server options")
    preset.add_argument(
        '--no-supports-http-options',
        dest='supports_http_options',
        action="store_false",
        help=" The command does not support project's HTTP server options")
    preset.add_argument('command', metavar="COMMAND", help="Command line or app filename to add")
    preset.set_defaults(main=command_commands.main, supports_http_options=None)

    preset = subparsers.add_parser('remove-command', help="Remove a command from the project")
    add_directory_arg(preset)
    add_command_name_arg(preset)
    preset.set_defaults(main=command_commands.main_remove)

    preset = subparsers.add_parser('list-commands', help="List the commands on the project")
    add_directory_arg(preset)
    preset.set_defaults(main=command_commands.main_list)

    # argparse doesn't do this for us for whatever reason
    if len(argv) < 2:
        print("Must specify a subcommand.", file=sys.stderr)
        parser.print_usage(file=sys.stderr)
        return 2  # argparse exits with 2 on bad args, copy that

    try:
        args = parser.parse_args(argv[1:])
    except SystemExit as e:
        return e.code

    if args.verbose:
        logger = (logging.getLoggerClass())(name="anaconda_project_verbose")
        logger.setLevel(logging.DEBUG)
        handler = logging.StreamHandler(stream=sys.stderr)
        logger.addHandler(handler)
        push_verbose_logger(logger)

    try:
        # '--directory' is used for most subcommands; for unarchive,
        # args.directory is positional and may be None
        if 'directory' in args and args.directory is not None:
            args.directory = os.path.abspath(args.directory)
        return args.main(args)
    finally:
        if args.verbose:
            pop_verbose_logger()


def _main_without_bug_handler():
    anaconda_project._enter_beta_test_mode()
    return _parse_args_and_run_subcommand(sys.argv)


def main():
    """anaconda-project command line tool Conda-style entry point.

    Conda expects us to take no args and return an exit code.
    """
    details = {'version': version}
    return handle_bugs(_main_without_bug_handler, program_name='anaconda-project', details_dict=details)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

from anaconda_project.conda_manager import (push_conda_manager_class, pop_conda_manager_class, new_conda_manager,
                                            CondaManager, CondaLockSet)
import anaconda_project.internal.conda_api as conda_api
from anaconda_project.yaml_file import _dump_string


def test_use_non_default_conda_manager():
    called = dict()

    class MyCondaManager(CondaManager):
        def __init__(self, frontend):
            pass

        def resolve_dependencies(self, package_specs, channels, platforms):
            return CondaLockSet({}, platforms=[])

        def find_environment_deviations(self, *args):
            called['find_environment_deviations'] = args

        def fix_environment_deviations(self, *args):
            called['fix_environment_deviations'] = args

        def remove_packages(self, *args):
            called['remove_packages'] = args

    push_conda_manager_class(MyCondaManager)
    try:
        manager = new_conda_manager()
        manager.find_environment_deviations(None, None)
        manager.fix_environment_deviations(None, None)
        manager.remove_packages(None, None)
        assert dict(
            find_environment_deviations=(None, None),
            fix_environment_deviations=(None, None),
            remove_packages=(None, None)) == called
    finally:
        pop_conda_manager_class()


def test_lock_set_properties(monkeypatch):
    lock_set = CondaLockSet({
        'all': ["something=0.5=2", "bokeh=0.12.4=1"],
        'linux-64': ["linux-thing=1.0=0"],
        'unix': ["unix-thing=5=1"],
        'win': ["windows-cross-bit-thing=3.2"],
        'win-32': ["windows-thing=2.0=3", "bokeh=2.3=7"]
    },
                            platforms=['linux-64', 'win-32'])
    # it is part of the API definition that we need to APPEND the
    # per-platform stuff, so it overrides.
    assert lock_set.package_specs_for_platform('win-32') == ("something=0.5=2", "windows-cross-bit-thing=3.2",
                                                             "windows-thing=2.0=3", "bokeh=2.3=7")

    # on Linux-64, test that it works without monkeypatch
    if conda_api.current_platform() != 'linux-64':
        monkeypatch.setattr('anaconda_project.internal.conda_api.current_platform', lambda: 'linux-64')

    assert lock_set.package_specs_for_current_platform == ("something=0.5=2", "bokeh=0.12.4=1", "unix-thing=5=1",
                                                           "linux-thing=1.0=0")

    assert lock_set.platforms == ('linux-64', 'win-32')


def test_lock_set_to_json(monkeypatch):
    lock_set = CondaLockSet({
        'all': ["something=0.5=2", "bokeh=0.12.4=1"],
        'linux-64': ["linux-thing=1.0=0"],
        'win-32': ["windows-thing=2.0=3", "bokeh=2.3=7"]
    },
                            platforms=['linux-64', 'win-32'])
    assert {
        'locked': True,
        'packages': {
            'all': ['something=0.5=2', 'bokeh=0.12.4=1'],
            'linux-64': ['linux-thing=1.0=0'],
            'win-32': ['windows-thing=2.0=3', 'bokeh=2.3=7']
        },
        'platforms': ['linux-64', 'win-32']
    } == lock_set.to_json()


def test_lock_set_to_yaml(monkeypatch):
    lock_set = CondaLockSet({
        'all': ['a', 'b'],
        'linux': ['x'],
        'win': ['y'],
        'linux-64': ['z', 'q'],
        'osx-64': ['s']
    },
                            platforms=['linux-64', 'win-64', 'osx-64'])

    # Mostly our interest here is that the ordering of the dict
    # is deterministic
    j = lock_set.to_json()
    assert _dump_string(j) == """locked: true
platforms:
- linux-64
- osx-64
- win-64
packages:
  all:
  - a
  - b
  linux:
  - x
  win:
  - y
  linux-64:
  - z
  - q
  osx-64:
  - s
"""


def test_lock_set_diff_and_equivalent():
    old_lock_set = CondaLockSet({
        'all': ['a', 'b'],
        'linux': ['x'],
        'win': ['y'],
        'linux-64': ['z', 'q'],
        'osx-64': ['s']
    },
                                platforms=['linux-64', 'osx-64'])
    new_lock_set = CondaLockSet({
        'all': ['a', 'b', 'c'],
        'linux': ['x', 'h'],
        'win': ['y'],
        'linux-64': ['q', 'w'],
        'osx-64': ['s'],
        'win-64': ['j']
    },
                                platforms=['linux-64', 'win-64'])

    assert """  platforms:
-   osx-64
+   win-64
  packages:
    all:
+     c
    linux:
+     h
    linux-64:
-     z
+     w
+   win-64:
+     j""" == new_lock_set.diff_from(old_lock_set)

    assert """  platforms:
-   win-64
+   osx-64
  packages:
    all:
-     c
    linux:
-     h
    linux-64:
+     z
-     w
-   win-64:
-     j""" == old_lock_set.diff_from(new_lock_set)

    assert "" == new_lock_set.diff_from(new_lock_set)
    assert "" == old_lock_set.diff_from(old_lock_set)

    assert old_lock_set.equivalent_to(old_lock_set)
    assert new_lock_set.equivalent_to(new_lock_set)
    assert not old_lock_set.equivalent_to(new_lock_set)
    assert not new_lock_set.equivalent_to(old_lock_set)

    assert """  platforms:
+   linux-64
+   win-64
  packages:
+   all:
+     a
+     b
+     c
+   linux:
+     x
+     h
+   win:
+     y
+   linux-64:
+     q
+     w
+   osx-64:
+     s
+   win-64:
+     j""" == new_lock_set.diff_from(None)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function
from functools import partial

import os

import anaconda_project
from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand

all_subcommands = ('init', 'run', 'prepare', 'clean', 'activate', 'archive', 'unarchive', 'upload', 'download',
                   'add-variable', 'remove-variable', 'list-variables', 'set-variable', 'unset-variable',
                   'add-download', 'remove-download', 'list-downloads', 'add-service', 'remove-service',
                   'list-services', 'add-env-spec', 'remove-env-spec', 'list-env-specs', 'export-env-spec', 'lock',
                   'unlock', 'update', 'add-packages', 'remove-packages', 'list-packages', 'add-platforms',
                   'remove-platforms', 'list-platforms', 'add-command', 'remove-command', 'list-commands')
all_subcommands_in_curlies = "{" + ",".join(all_subcommands) + "}"
all_subcommands_comma_space = ", ".join(["'" + s + "'" for s in all_subcommands])


def test_main_no_subcommand(capsys):
    code = _parse_args_and_run_subcommand(['project'])

    assert 2 == code

    out, err = capsys.readouterr()
    assert "" == out
    expected_error_msg = ('Must specify a subcommand.\n'
                          'usage: anaconda-project [-h] [-v] [--verbose]\n'
                          '                        %s\n'
                          '                        ...\n') % all_subcommands_in_curlies
    assert expected_error_msg == err


def test_main_bad_subcommand(capsys):
    code = _parse_args_and_run_subcommand(['project', 'foo'])

    out, err = capsys.readouterr()
    expected_error_msg = ("usage: anaconda-project [-h] [-v] [--verbose]\n"
                          "                        %s\n"
                          "                        ...\nanaconda-project: error: invalid choice: 'foo' "
                          "(choose from %s)\n") % (all_subcommands_in_curlies, all_subcommands_comma_space)
    assert expected_error_msg == err
    assert "" == out

    assert 2 == code


expected_usage_msg_format = (  # noqa
    'usage: anaconda-project [-h] [-v] [--verbose]\n'
    '                        %s\n'
    '                        ...\n'
    '\n'
    'Actions on projects (runnable projects).\n'
    '\n'
    'positional arguments:\n'
    '  %s\n'
    '                        Sub-commands\n'
    '    init                Initialize a directory with default project\n'
    '                        configuration\n'
    '    run                 Run the project, setting up requirements first\n'
    '    prepare             Set up the project requirements, but does not run the\n'
    '                        project\n'
    '    clean               Removes generated state (stops services, deletes\n'
    '                        environment files, etc)\n'
    '%s'
    '    archive             Create a .zip, .tar.gz, or .tar.bz2 archive with\n'
    '                        project files in it\n'
    '    unarchive           Unpack a .zip, .tar.gz, or .tar.bz2 archive with\n'
    '                        project files in it\n'
    '    upload              Upload the project to Anaconda Cloud\n'
    '    download            Download the project from Anaconda Cloud\n'
    '    add-variable        Add a required environment variable to the project\n'
    '    remove-variable     Remove an environment variable from the project\n'
    '    list-variables      List all variables on the project\n'
    '    set-variable        Set an environment variable value in anaconda-project-\n'
    '                        local.yml\n'
    '    unset-variable      Unset an environment variable value from anaconda-\n'
    '                        project-local.yml\n'
    '    add-download        Add a URL to be downloaded before running commands\n'
    '    remove-download     Remove a download from the project and from the\n'
    '                        filesystem\n'
    '    list-downloads      List all downloads on the project\n'
    '    add-service         Add a service to be available before running commands\n'
    '    remove-service      Remove a service from the project\n'
    '    list-services       List services present in the project\n'
    '    add-env-spec        Add a new environment spec to the project\n'
    '    remove-env-spec     Remove an environment spec from the project\n'
    '    list-env-specs      List all environment specs for the project\n'
    '    export-env-spec     Save an environment spec as a conda environment file\n'
    '    lock                Lock all packages at their current versions\n'
    '    unlock              Remove locked package versions\n'
    '    update              Update all packages to their latest versions\n'
    '    add-packages        Add packages to one or all project environments\n'
    '    remove-packages     Remove packages from one or all project environments\n'
    '    list-packages       List packages for an environment on the project\n'
    '    add-platforms       Add platforms to one or all project environments\n'
    '    remove-platforms    Remove platforms from one or all project environments\n'
    '    list-platforms      List platforms for an environment on the project\n'
    '    add-command         Add a new command to the project\n'
    '    remove-command      Remove a command from the project\n'
    '    list-commands       List the commands on the project\n'
    '\n'
    'optional arguments:\n'
    '  -h, --help            show this help message and exit\n'
    "  -v, --version         show program's version number and exit\n"
    '  --verbose             show verbose debugging details\n')

activate_help = ('    activate            Set up the project and output shell export commands\n'
                 '                        reflecting the setup\n')

expected_usage_msg = expected_usage_msg_format % (all_subcommands_in_curlies, all_subcommands_in_curlies, activate_help)

expected_usage_msg_without_activate = expected_usage_msg_format % (all_subcommands_in_curlies.replace(
    ",activate", ""), all_subcommands_in_curlies.replace(",activate", ""), "")


def test_main_help(capsys):
    code = _parse_args_and_run_subcommand(['project', '--help'])

    out, err = capsys.readouterr()

    assert "" == err
    assert expected_usage_msg == out

    assert 0 == code


def test_main_help_via_entry_point(capsys, monkeypatch):
    from anaconda_project.internal.cli.main import main

    monkeypatch.setattr("sys.argv", ['project', '--help'])

    code = main()

    out, err = capsys.readouterr()

    assert "" == err
    assert expected_usage_msg_without_activate == out

    assert 0 == code

    # undo this side effect
    anaconda_project._beta_test_mode = False


def _main_calls_subcommand(monkeypatch, capsys, subcommand):
    def mock_subcommand_main(subcommand, args):
        print("Hi I am subcommand {}".format(subcommand))
        assert args.directory == os.path.abspath('MYPROJECT')
        return 27

    monkeypatch.setattr('anaconda_project.internal.cli.{}.main'.format(subcommand),
                        partial(mock_subcommand_main, subcommand))
    code = _parse_args_and_run_subcommand(['anaconda-project', subcommand, '--directory', 'MYPROJECT'])

    assert 27 == code

    out, err = capsys.readouterr()
    assert ("Hi I am subcommand {}\n".format(subcommand)) == out
    assert "" == err


def test_main_calls_run(monkeypatch, capsys):
    _main_calls_subcommand(monkeypatch, capsys, 'run')


def test_main_calls_prepare(monkeypatch, capsys):
    _main_calls_subcommand(monkeypatch, capsys, 'prepare')


def test_main_when_buggy(capsys, monkeypatch):
    from anaconda_project.internal.cli.main import main

    def mock_is_interactive():
        return True

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_is_interactive)

    def mock_main():
        raise AssertionError("It did not work")

    monkeypatch.setattr('anaconda_project.internal.cli.main._main_without_bug_handler', mock_main)
    monkeypatch.setattr("sys.argv", ['anaconda-project'])

    result = main()
    assert result == 1
    out, err = capsys.readouterr()

    assert '' == out
    assert err.startswith("""An unexpected error occurred, most likely a bug in anaconda-project.
    (The error was: AssertionError: It did not work)
Details about the error were saved to """)
    filename = err.split()[-1]
    assert filename.endswith(".txt")
    assert os.path.basename(filename).startswith("bug_details_anaconda-project_")
    assert os.path.isfile(filename)

    os.remove(filename)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import anaconda_project.cli as cli


def test_main(monkeypatch):
    result = {}

    def mock_main(*args, **kwargs):
        result['args'] = args
        result['kwargs'] = kwargs

    monkeypatch.setattr('anaconda_project.internal.cli.main.main', mock_main)
    cli.main()

    assert dict(args=(), kwargs={}) == result



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Commands related to setting and unsetting variables."""
from __future__ import absolute_import, print_function

import sys

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project import project_ops
from anaconda_project.internal.cli import console_utils


def add_variables(project_dir, env_spec_name, vars_to_add, default):
    """Add env variables to project file.

    Returns:
        Returns exit code
    """
    if len(vars_to_add) > 1 and default is not None:
        # yapf on P2.7 vs yapf on P3.6 disagree here.
        # yapf: disable
        print("It isn't clear which variable your --default option goes with; "
              "add one variable at a time if using --default.", file=sys.stderr)
        # yapf: enable
        return 1
    project = load_project(project_dir)
    status = project_ops.add_variables(project, env_spec_name, vars_to_add, {vars_to_add[0]: default})
    if status:
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def remove_variables(project_dir, env_spec_name, vars_to_remove):
    """Remove env variable requirements from the project file.

    Returns:
        Returns exit code
    """
    project = load_project(project_dir)
    status = project_ops.remove_variables(project, env_spec_name, vars_to_remove)
    if status:
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def list_variables(project_dir, env_spec_name):
    """List variables present in project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    print("Variables for project: {}\n".format(project_dir))
    console_utils.print_names_and_descriptions(project.all_variable_requirements(env_spec_name), name_attr='env_var')
    return 0


def set_variables(project_dir, env_spec_name, vars_and_values):
    """Set the given variables to the given values.

    Returns:
        Returns exit code
    """
    fixed_vars = []
    for var in vars_and_values:
        if '=' not in var:
            print("Error: argument '{}' should be in NAME=value format".format(var))
            return 1
        # maxsplit=1 -- no maxsplit keywork in py27
        fixed_vars.append(tuple(var.split('=', 1)))
    project = load_project(project_dir)
    status = project_ops.set_variables(project, env_spec_name, fixed_vars)
    if status:
        print(status.status_description)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def unset_variables(project_dir, env_spec_name, vars_to_unset):
    """Unset the variables for local project.

    Returns:
        Returns exit code
    """
    project = load_project(project_dir)
    status = project_ops.unset_variables(project, env_spec_name, vars_to_unset)
    if status:
        print(status.status_description)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def main_add(args):
    """Add variables main."""
    return add_variables(args.directory, args.env_spec, args.vars_to_add, args.default)


def main_remove(args):
    """Remove variables main."""
    return remove_variables(args.directory, args.env_spec, args.vars_to_remove)


def main_list(args):
    """List the project variable names."""
    return list_variables(args.directory, args.env_spec)


def main_set(args):
    """Set the project variables."""
    return set_variables(args.directory, args.env_spec, args.vars_and_values)


def main_unset(args):
    """Unset the project variables."""
    return unset_variables(args.directory, args.env_spec, args.vars_to_unset)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import shutil
import tempfile
import zipfile

from anaconda_project.internal import rename


# we overwrite as long as the zip contains a file and target_path
# is a file, or the zip is a dir and target_path is a dir, but if
# they don't match we don't overwrite. Hopefully this will catch
# most mistaken collisions.
def unpack_zip(zip_path, target_path, errors):
    try:
        with zipfile.ZipFile(zip_path, mode='r') as zf:
            target_dir, target_file = os.path.split(target_path)
            tmp_dir = tempfile.mkdtemp(prefix=(target_path + "_tmp"), dir=target_dir)
            try:
                zf.extractall(tmp_dir)
                extracted = os.listdir(tmp_dir)
                if len(extracted) == 0:
                    errors.append("Zip archive was empty.")
                    return False
                elif len(extracted) == 1 and extracted[0] == target_file:
                    # don't keep a pointless directory level, if
                    # the zip just contains a single directory or
                    # file with the same name as the target
                    src_path = os.path.join(tmp_dir, extracted[0])
                else:
                    src_path = tmp_dir
                src_is_dir = os.path.isdir(src_path)
                target_is_dir = os.path.isdir(target_path)
                if os.path.exists(target_path) and (src_is_dir != target_is_dir):
                    if src_is_dir:
                        errors.append(
                            "%s exists and isn't a directory, not unzipping a directory over it." % target_path)
                    else:
                        errors.append("%s exists and is a directory, not unzipping a plain file over it." % target_path)
                    return False
                else:
                    rename.rename_over_existing(src_path, target_path)
            finally:
                if os.path.isdir(tmp_dir):
                    shutil.rmtree(path=tmp_dir)
            return True
    except Exception as e:
        errors.append("Failed to unzip %s: %s" % (zip_path, str(e)))
        return False
